import logging
import typing
from datetime import datetime
from decimal import Decimal
from typing import Any, List, Literal, Optional, Union

import numpy
import xarray
from dateutil.tz import UTC

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineCoordinate,
    CopernicusMarinePart,
    CopernicusMarineService,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.exceptions import (
    CoordinatesOutOfDatasetBounds,
    MinimumLongitudeGreaterThanMaximumLongitude,
    ServiceNotSupported,
    VariableDoesNotExistInTheDataset,
)
from copernicusmarine.core_functions.models import CoordinatesSelectionMethod
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.utils import (
    timestamp_or_datestring_to_datetime,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    TemporalParameters,
    XParameters,
    YParameters,
)
from copernicusmarine.download_functions.utils import (
    get_coordinate_ids_from_parameters,
)

logger = logging.getLogger("copernicusmarine")


NETCDF_CONVENTION_VARIABLE_ATTRIBUTES = [
    "standard_name",
    "long_name",
    "units",
    "unit_long",
    "valid_min",
    "valid_max",
]
NETCDF_CONVENTION_COORDINATE_ATTRIBUTES = [
    "standard_name",
    "long_name",
    "units",
    "unit_long",
    "axis",
]
NETCDF_CONVENTION_DATASET_ATTRIBUTES = [
    "title",
    "institution",
    "source",
    "history",
    "references",
    "comment",
    "Conventions",
    "producer",
    "credit",
    "contact",
]


@typing.no_type_check
def _choose_extreme_point(
    dataset: xarray.Dataset,
    coord_label: str,
    actual_extreme: Union[float, datetime],
    method: Literal["pad", "backfill", "nearest"],
) -> Union[float, datetime]:
    if (
        coord_label == "time"
        and actual_extreme
        >= timestamp_or_datestring_to_datetime(
            dataset[coord_label].values.min()
        ).replace(tzinfo=None)
        and actual_extreme
        <= timestamp_or_datestring_to_datetime(
            dataset[coord_label].values.max()
        ).replace(tzinfo=None)
    ):
        external_point = dataset.sel(
            {coord_label: actual_extreme}, method=method
        )[coord_label].values
        external_point = timestamp_or_datestring_to_datetime(
            external_point
        ).replace(tzinfo=None)
    elif (
        coord_label != "time"
        and actual_extreme > dataset[coord_label].min()
        and method == "nearest"
    ):
        external_point = dataset.sel(
            {coord_label: actual_extreme}, method=method
        )[coord_label].values
    elif (
        coord_label != "time"
        and actual_extreme > dataset[coord_label].min()
        and actual_extreme < dataset[coord_label].max()
    ):
        external_point = dataset.sel(
            {coord_label: actual_extreme}, method=method
        )[coord_label].values
    else:
        external_point = actual_extreme
    return external_point


def _enlarge_selection(
    dataset: xarray.Dataset,
    coord_label: str,
    coord_selection: slice,
) -> slice:
    external_minimum = _choose_extreme_point(
        dataset, coord_label, coord_selection.start, "pad"
    )

    external_maximum = _choose_extreme_point(
        dataset, coord_label, coord_selection.stop, "backfill"
    )

    return slice(external_minimum, external_maximum)


def _nearest_selection(
    dataset: xarray.Dataset,
    coord_label: str,
    coord_selection: slice,
) -> slice:
    external_minimum = _choose_extreme_point(
        dataset, coord_label, coord_selection.start, "nearest"
    )

    external_maximum = _choose_extreme_point(
        dataset, coord_label, coord_selection.stop, "nearest"
    )

    return slice(external_minimum, external_maximum)


def _dataset_custom_sel(
    dataset: xarray.Dataset,
    coordinate_label: str,
    coord_selection: Union[float, slice, datetime, None],
    coordinates_selection_method: CoordinatesSelectionMethod,
) -> xarray.Dataset:
    if coordinate_label in dataset.sizes:
        if isinstance(coord_selection, slice):
            if (
                len(dataset[coordinate_label].values) > 1
                and (
                    dataset[coordinate_label].values[0]
                    > dataset[coordinate_label].values[1]
                )
                and coord_selection.start < coord_selection.stop
            ):
                coord_selection = slice(
                    coord_selection.stop, coord_selection.start
                )
        if coordinates_selection_method == "outside":
            if (
                isinstance(coord_selection, slice)
                and coord_selection.stop is not None
            ):
                coord_selection = _enlarge_selection(
                    dataset, coordinate_label, coord_selection
                )
        if coordinates_selection_method == "nearest":
            if (
                isinstance(coord_selection, slice)
                and coord_selection.stop is not None
            ):
                coord_selection = _nearest_selection(
                    dataset, coordinate_label, coord_selection
                )
        if isinstance(coord_selection, slice):
            tmp_dataset = dataset.sel(
                {coordinate_label: coord_selection}, method=None
            )
        else:
            tmp_dataset = dataset.sel(
                {coordinate_label: coord_selection}, method="nearest"
            )
        if tmp_dataset.coords[coordinate_label].size == 0 or (
            coordinate_label not in tmp_dataset.sizes
        ):
            target = (
                coord_selection.start
                if isinstance(coord_selection, slice)
                else coord_selection
            )
            nearest_neighbour_value = dataset.sel(
                {coordinate_label: target}, method="nearest"
            )[coordinate_label].values
            dataset = dataset.sel(
                {
                    coordinate_label: slice(
                        nearest_neighbour_value, nearest_neighbour_value
                    )
                }
            )
        else:
            dataset = tmp_dataset
    return dataset


def get_size_of_coordinate_subset(
    dataset: xarray.Dataset,
    coordinate_label: str,
    minimum: Optional[Union[float, datetime]],
    maximum: Optional[Union[float, datetime]],
) -> int:
    if coordinate_label in dataset.sizes:
        return (
            dataset.coords[coordinate_label]
            .sel({coordinate_label: slice(minimum, maximum)}, method=None)
            .coords[coordinate_label]
            .size
        )
    else:
        raise KeyError(
            f"Could not subset on {coordinate_label}. "
            "Didn't find an equivalent in the dataset."
        )


def _shift_longitude_dimension(
    dataset: xarray.Dataset,
    minimum_longitude_modulus: float,
    coordinate_id: str,
    coordinates_selection_method: CoordinatesSelectionMethod,
):
    if coordinates_selection_method == "outside":
        minimum_longitude_modulus = _choose_extreme_point(
            dataset,
            coordinate_id,
            minimum_longitude_modulus,
            "pad",
        )  # type: ignore
    if coordinates_selection_method == "nearest":
        minimum_longitude_modulus = _choose_extreme_point(
            dataset,
            coordinate_id,
            minimum_longitude_modulus,
            "nearest",
        )  # type: ignore
    window = (
        minimum_longitude_modulus + 180
    )  # compute the degrees needed to move the dataset
    if coordinate_id in dataset.sizes:
        attrs = dataset[coordinate_id].attrs
        if "valid_min" in attrs:
            attrs["valid_min"] += window
        if "valid_max" in attrs:
            attrs["valid_max"] += window
        dataset = dataset.assign_coords(
            {
                coordinate_id: (
                    (dataset[coordinate_id] + (180 - window)) % 360
                )
                - (180 - window)
            }
        ).sortby(coordinate_id)
        dataset[coordinate_id].attrs = attrs
    return dataset


def _y_axis_subset(
    dataset: xarray.Dataset,
    y_parameters: YParameters,
    coordinates_selection_method: CoordinatesSelectionMethod,
) -> xarray.Dataset:
    minimum_y = y_parameters.minimum_y
    maximum_y = y_parameters.maximum_y
    if minimum_y is not None or maximum_y is not None:
        y_selection = (
            minimum_y
            if minimum_y == maximum_y
            else slice(minimum_y, maximum_y)
        )
        return _dataset_custom_sel(
            dataset,
            y_parameters.coordinate_id,
            y_selection,
            coordinates_selection_method,
        )

    return dataset


def x_axis_selection(
    longitude_parameters: XParameters,
) -> tuple[Union[float, slice, None], bool]:
    shift_window = False
    minimum_x = longitude_parameters.minimum_x
    maximum_x = longitude_parameters.maximum_x
    if minimum_x is not None and maximum_x is not None:
        if longitude_parameters.coordinate_id == "longitude":
            if minimum_x > maximum_x:
                raise MinimumLongitudeGreaterThanMaximumLongitude(
                    "--minimum-longitude option must be smaller "
                    "or equal to --maximum-longitude"
                )
            if maximum_x - minimum_x >= 360:
                if maximum_x != 180:
                    shift_window = True
                return slice(-180, 180), shift_window
            else:
                minimum_x = longitude_modulus(minimum_x)
                maximum_x = longitude_modulus(maximum_x)

            if maximum_x and minimum_x is not None and maximum_x < minimum_x:
                maximum_x += 360
                shift_window = True

        return (
            minimum_x
            if minimum_x == maximum_x
            else slice(minimum_x, maximum_x)
        ), shift_window
    if minimum_x is not None:
        return slice(minimum_x, None), shift_window
    elif maximum_x is not None:
        return slice(None, maximum_x), shift_window

    return None, shift_window


def _x_axis_subset(
    dataset: xarray.Dataset,
    longitude_parameters: XParameters,
    coordinates_selection_method: CoordinatesSelectionMethod,
) -> xarray.Dataset:
    (x_selection, shift_window) = x_axis_selection(longitude_parameters)
    if shift_window and isinstance(x_selection, slice):
        dataset = _shift_longitude_dimension(
            dataset,
            x_selection.start,
            longitude_parameters.coordinate_id,
            coordinates_selection_method,
        )

    if x_selection is not None:

        return _dataset_custom_sel(
            dataset,
            longitude_parameters.coordinate_id,
            x_selection,
            coordinates_selection_method,
        )
    return dataset


def t_axis_selection(
    temporal_parameters: TemporalParameters,
) -> Union[slice, datetime, None]:

    start_datetime = (
        temporal_parameters.start_datetime.astimezone(UTC).replace(tzinfo=None)
        if temporal_parameters.start_datetime
        else temporal_parameters.start_datetime
    )
    end_datetime = (
        temporal_parameters.end_datetime.astimezone(UTC).replace(tzinfo=None)
        if temporal_parameters.end_datetime
        else temporal_parameters.end_datetime
    )

    if start_datetime is not None or end_datetime is not None:
        return (
            start_datetime
            if start_datetime == end_datetime
            else slice(start_datetime, end_datetime)
        )
    return None


def _temporal_subset(
    dataset: xarray.Dataset,
    temporal_parameters: TemporalParameters,
    coordinates_selection_method: CoordinatesSelectionMethod,
) -> xarray.Dataset:
    temporal_selection = t_axis_selection(temporal_parameters)
    if temporal_selection is not None:
        dataset = _dataset_custom_sel(
            dataset,
            "time",
            temporal_selection,
            coordinates_selection_method,
        )
    return dataset


def _depth_subset(
    dataset: xarray.Dataset,
    depth_parameters: DepthParameters,
    coordinates_selection_method: CoordinatesSelectionMethod,
) -> xarray.Dataset:
    def convert_elevation_to_depth(dataset: xarray.Dataset):
        if "elevation" in dataset.sizes:
            attrs = dataset["elevation"].attrs
            dataset = dataset.reindex(elevation=dataset.elevation[::-1])
            dataset["elevation"] = dataset.elevation * (-1)
            attrs["positive"] = "down"
            attrs["standard_name"] = "depth"
            attrs["long_name"] = "Depth"
            attrs["units"] = "m"
            dataset = dataset.rename({"elevation": "depth"})
            dataset.depth.attrs = attrs
        return dataset

    def update_elevation_attributes(dataset: xarray.Dataset):
        if "elevation" in dataset.sizes:
            attrs = dataset["elevation"].attrs
            attrs["positive"] = "up"
            attrs["standard_name"] = "elevation"
            attrs["long_name"] = "Elevation"
            attrs["units"] = "m"
            dataset["elevation"].attrs = attrs
        return dataset

    if depth_parameters.vertical_axis == "depth":
        dataset = convert_elevation_to_depth(dataset)
    else:
        dataset = update_elevation_attributes(dataset)
    minimum_depth = depth_parameters.minimum_depth
    maximum_depth = depth_parameters.maximum_depth
    if minimum_depth is not None or maximum_depth is not None:
        coords = (
            dataset.coords if isinstance(dataset, xarray.Dataset) else dataset
        )
        if "elevation" in coords:
            minimum_depth = (
                minimum_depth * -1.0 if minimum_depth is not None else None
            )
            maximum_depth = (
                maximum_depth * -1.0 if maximum_depth is not None else None
            )
            minimum_depth, maximum_depth = maximum_depth, minimum_depth

        depth_selection = (
            minimum_depth
            if minimum_depth == maximum_depth
            else slice(minimum_depth, maximum_depth)
        )
        dataset = _dataset_custom_sel(
            dataset,
            depth_parameters.vertical_axis,
            depth_selection,
            coordinates_selection_method,
        )
    return dataset


def _get_variable_name_from_standard_name(
    dataset: xarray.Dataset, standard_name: str
) -> Optional[str]:
    for variable_name in dataset.variables:
        if (
            hasattr(dataset[variable_name], "standard_name")
            and dataset[variable_name].standard_name == standard_name
        ):
            return str(variable_name)
    return None


def _adequate_dtypes_of_valid_minmax(
    dataset: xarray.Dataset, variable: str
) -> xarray.Dataset:
    dataset[variable].attrs["valid_min"] = numpy.array(
        [dataset[variable].attrs["valid_min"]],
        dtype=dataset[variable].encoding["dtype"],
    )[0]
    dataset[variable].attrs["valid_max"] = numpy.array(
        [dataset[variable].attrs["valid_max"]],
        dtype=dataset[variable].encoding["dtype"],
    )[0]
    return dataset


def _update_variables_attributes(
    dataset: xarray.Dataset, variables: List[str]
) -> xarray.Dataset:
    for variable in variables:
        dataset[variable].attrs = _filter_attributes(
            dataset[variable].attrs, NETCDF_CONVENTION_VARIABLE_ATTRIBUTES
        )
        if (
            "valid_min" in dataset[variable].attrs
            and "valid_max" in dataset[variable].attrs
        ):
            _adequate_dtypes_of_valid_minmax(dataset, variable)
    return dataset


def _variables_subset(
    dataset: xarray.Dataset, variables: List[str]
) -> xarray.Dataset:
    dataset_variables_filter = []

    for variable in variables:
        if variable in dataset.variables:
            dataset_variables_filter.append(variable)
        else:
            variable_name_from_standard_name = (
                _get_variable_name_from_standard_name(dataset, variable)
            )
            if variable_name_from_standard_name is not None:
                dataset_variables_filter.append(
                    variable_name_from_standard_name
                )
            else:
                raise VariableDoesNotExistInTheDataset(variable)
    dataset = dataset[numpy.array(dataset_variables_filter)]
    return _update_variables_attributes(dataset, dataset_variables_filter)


def _filter_attributes(attributes: dict, attributes_to_keep: List[str]):
    attributes_that_exist = set(attributes).intersection(attributes_to_keep)
    return {key: attributes[key] for key in attributes_that_exist}


def _update_dataset_coordinate_attributes(
    dataset: xarray.Dataset,
    coordinate_ids: list[str],
) -> xarray.Dataset:
    for coordinate_id in coordinate_ids:
        if coordinate_id in dataset.sizes:
            coord = dataset[coordinate_id]
            attrs = coord.attrs
            coordinate_attributes = (
                NETCDF_CONVENTION_COORDINATE_ATTRIBUTES.copy()
            )
            if "time" in coordinate_id:
                attrs["standard_name"] = "time"
                attrs["long_name"] = "Time"
                attrs["axis"] = "T"
                attrs["unit_long"] = (
                    coord.encoding["units"].replace("_", " ").title()
                )
                coordinate_attributes.remove("units")
            elif coordinate_id in ["depth", "elevation"]:
                coordinate_attributes.append("positive")
            # TODO: delete this when fixed on ARCO processor side
            # example dataset: esa_obs-si_arc_phy-sit_nrt_l4-multi_P1D-m
            if attrs == {}:
                if coordinate_id == "longitude":
                    attrs = {
                        "standard_name": "longitude",
                        "long_name": "Longitude",
                        "units": "degrees_east",
                        "units_long": "Degrees East",
                        "axis": "X",
                    }
                elif coordinate_id == "latitude":
                    attrs = {
                        "standard_name": "latitude",
                        "long_name": "Latitude",
                        "units": "degrees_north",
                        "units_long": "Degrees North",
                        "axis": "Y",
                    }
            coord.attrs = _filter_attributes(attrs, coordinate_attributes)

    dataset.attrs = _filter_attributes(
        dataset.attrs, NETCDF_CONVENTION_DATASET_ATTRIBUTES
    )

    return dataset


def subset(
    dataset: xarray.Dataset,
    variables: Optional[List[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
    coordinates_selection_method: CoordinatesSelectionMethod,
) -> xarray.Dataset:
    if variables:
        dataset = _variables_subset(dataset, variables)
    dataset = _y_axis_subset(
        dataset,
        geographical_parameters.y_axis_parameters,
        coordinates_selection_method,
    )
    dataset = _x_axis_subset(
        dataset,
        geographical_parameters.x_axis_parameters,
        coordinates_selection_method,
    )

    dataset = _temporal_subset(
        dataset, temporal_parameters, coordinates_selection_method
    )

    dataset = _depth_subset(
        dataset, depth_parameters, coordinates_selection_method
    )

    dataset = _update_dataset_coordinate_attributes(
        dataset,
        get_coordinate_ids_from_parameters(
            geographical_parameters, temporal_parameters, depth_parameters
        ),
    )

    return dataset


def longitude_modulus(longitude: float) -> float:
    """
    Returns the equivalent longitude in [-180, 180[
    """
    # We are using Decimal to avoid issue with rounding
    modulus = (Decimal(str(longitude)) + 180) % 360
    # Modulus with python return a negative value if the denominator is negative
    # To counteract that, we add 360 if the result is < 0
    modulus = modulus if modulus >= 0 else modulus + 360
    return float(modulus - 180)


def longitude_modulus_upper_bound(longitude: float) -> float:
    """
    Returns the equivalent longitude in ]-180, 180]
    """
    modulus = longitude_modulus(longitude)
    if modulus == -180:
        return 180.0
    return modulus


def check_dataset_subset_bounds(
    service: CopernicusMarineService,
    part: CopernicusMarinePart,
    dataset_subset: SubsetRequest,
    coordinates_selection_method: CoordinatesSelectionMethod,
    axis_coordinate_id_mapping: dict[str, str],
) -> None:
    # TODO: check lon/lat for original grid
    if service.service_name not in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
        CopernicusMarineServiceNames.PLATFORMSERIES,
    ]:
        raise ServiceNotSupported(service.service_name)
    all_coordinates = part.get_coordinates()
    if "y" in axis_coordinate_id_mapping:
        coordinate_id = axis_coordinate_id_mapping["y"]
        coordinate, _, _ = all_coordinates[coordinate_id]
        minimum_value, maximum_value = _get_minimun_maximum_dataset(coordinate)
        user_minimum_coordinate_value = (
            dataset_subset.minimum_y
            if dataset_subset.minimum_y is not None
            else minimum_value
        )
        user_maximum_coordinate_value = (
            dataset_subset.maximum_y
            if dataset_subset.maximum_y is not None
            else maximum_value
        )
        _check_coordinate_overlap(
            dimension=coordinate_id,
            user_minimum_coordinate_value=user_minimum_coordinate_value,
            user_maximum_coordinate_value=user_maximum_coordinate_value,
            dataset_minimum_coordinate_value=minimum_value,
            dataset_maximum_coordinate_value=maximum_value,
            is_strict=coordinates_selection_method == "strict-inside",
        )
    if "x" in axis_coordinate_id_mapping:
        coordinate_id = axis_coordinate_id_mapping["x"]
        coordinate, _, _ = all_coordinates[coordinate_id]
        minimum_value, maximum_value = _get_minimun_maximum_dataset(coordinate)
        if coordinate_id == "longitude":
            if (
                dataset_subset.minimum_x is not None
                and dataset_subset.maximum_x is not None
                and dataset_subset.minimum_x > dataset_subset.maximum_x
            ):
                raise MinimumLongitudeGreaterThanMaximumLongitude(
                    "--minimum-longitude option must be smaller "
                    "or equal to --maximum-longitude"
                )
            user_minimum_coordinate_value = (
                longitude_modulus(dataset_subset.minimum_x)
                if dataset_subset.minimum_x is not None
                else minimum_value
            )
            user_maximum_coordinate_value = (
                longitude_modulus_upper_bound(dataset_subset.maximum_x)
                if dataset_subset.maximum_x is not None
                else maximum_value
            )
        else:
            user_minimum_coordinate_value = (
                dataset_subset.minimum_x
                if dataset_subset.minimum_x is not None
                else minimum_value
            )
            user_maximum_coordinate_value = (
                dataset_subset.maximum_x
                if dataset_subset.maximum_x is not None
                else maximum_value
            )
        _check_coordinate_overlap(
            dimension=coordinate_id,
            user_minimum_coordinate_value=user_minimum_coordinate_value,
            user_maximum_coordinate_value=user_maximum_coordinate_value,
            dataset_minimum_coordinate_value=minimum_value,
            dataset_maximum_coordinate_value=maximum_value,
            is_strict=coordinates_selection_method == "strict-inside",
        )
    if "t" in axis_coordinate_id_mapping:
        coordinate_id = axis_coordinate_id_mapping["t"]
        coordinate, _, _ = all_coordinates[coordinate_id]
        minimum_value, maximum_value = _get_minimun_maximum_dataset(coordinate)
        dataset_minimum_coordinate_value = timestamp_or_datestring_to_datetime(
            minimum_value
        )
        dataset_maximum_coordinate_value = timestamp_or_datestring_to_datetime(
            maximum_value
        )
        user_minimum_coordinate_value = (
            dataset_subset.start_datetime
            if dataset_subset.start_datetime is not None
            else dataset_minimum_coordinate_value
        )
        user_maximum_coordinate_value = (
            dataset_subset.end_datetime
            if dataset_subset.end_datetime is not None
            else dataset_maximum_coordinate_value
        )
        _check_coordinate_overlap(
            dimension="time",
            user_minimum_coordinate_value=user_minimum_coordinate_value,
            user_maximum_coordinate_value=user_maximum_coordinate_value,
            dataset_minimum_coordinate_value=dataset_minimum_coordinate_value,
            dataset_maximum_coordinate_value=dataset_maximum_coordinate_value,
            is_strict=coordinates_selection_method == "strict-inside",
        )
    if "z" in axis_coordinate_id_mapping:
        coordinate_id = axis_coordinate_id_mapping["z"]
        coordinate, _, _ = all_coordinates[coordinate_id]
        minimum_value, maximum_value = _get_minimun_maximum_dataset(coordinate)
        _check_coordinate_overlap(
            dimension="depth",
            user_minimum_coordinate_value=(
                dataset_subset.minimum_depth
                if dataset_subset.minimum_depth is not None
                else minimum_value
            ),
            user_maximum_coordinate_value=(
                dataset_subset.maximum_depth
                if dataset_subset.maximum_depth is not None
                else maximum_value
            ),
            dataset_minimum_coordinate_value=minimum_value,
            dataset_maximum_coordinate_value=maximum_value,
            is_strict=coordinates_selection_method == "strict-inside",
        )


@typing.no_type_check
def _check_coordinate_overlap(
    dimension: str,
    user_minimum_coordinate_value: Union[float, datetime],
    user_maximum_coordinate_value: Union[float, datetime],
    dataset_minimum_coordinate_value: Union[float, datetime],
    dataset_maximum_coordinate_value: Union[float, datetime],
    is_strict: bool,
) -> None:
    message = (
        f"Some of your subset selection "
        f"[{user_minimum_coordinate_value}, {user_maximum_coordinate_value}] "
        f"for the {dimension} dimension exceed the dataset coordinates "
        f"[{dataset_minimum_coordinate_value}, "
        f"{dataset_maximum_coordinate_value}]"
    )
    if dataset_maximum_coordinate_value < dataset_minimum_coordinate_value:
        dataset_maximum_coordinate_value, dataset_minimum_coordinate_value = (
            dataset_minimum_coordinate_value,
            dataset_maximum_coordinate_value,
        )
    if dimension == "longitude":
        if dataset_minimum_coordinate_value == -180:
            dataset_maximum_coordinate_value = 180
        if dataset_maximum_coordinate_value == 180:
            dataset_minimum_coordinate_value = -180
    if user_maximum_coordinate_value < dataset_minimum_coordinate_value:
        if user_minimum_coordinate_value == dataset_minimum_coordinate_value:
            message = (
                f"Some of your subset selection "
                f"({dimension} < {user_maximum_coordinate_value}) "
                f"for the {dimension} dimension exceed the dataset coordinates "
                f"[{dataset_minimum_coordinate_value}, "
                f"{dataset_maximum_coordinate_value}]"
            )
        raise CoordinatesOutOfDatasetBounds(message)
    elif user_minimum_coordinate_value > dataset_maximum_coordinate_value:
        if user_maximum_coordinate_value == dataset_maximum_coordinate_value:
            message = (
                f"Some of your subset selection "
                f"({dimension} > {user_minimum_coordinate_value}) "
                f"for the {dimension} dimension exceed the dataset coordinates "
                f"[{dataset_minimum_coordinate_value}, "
                f"{dataset_maximum_coordinate_value}]"
            )
        raise CoordinatesOutOfDatasetBounds(message)
    elif (
        (
            user_minimum_coordinate_value < dataset_minimum_coordinate_value
            and user_maximum_coordinate_value
            > dataset_maximum_coordinate_value
        )
        or user_minimum_coordinate_value < dataset_minimum_coordinate_value
        or user_maximum_coordinate_value > dataset_maximum_coordinate_value
    ):
        if is_strict:
            raise CoordinatesOutOfDatasetBounds(message)
        else:
            logger.warning(message)


def _get_minimun_maximum_dataset(
    coordinate: CopernicusMarineCoordinate,
) -> tuple[Any, Any]:
    maximum_value = coordinate.maximum_value
    minimum_value = coordinate.minimum_value
    values = coordinate.values

    if maximum_value is None and values:
        maximum_value = max(values)
    if minimum_value is None and values:
        minimum_value = min(values)
    return minimum_value, maximum_value



import logging
import pathlib
from datetime import datetime
from typing import Any, Optional, Union

import xarray

from copernicusmarine.core_functions.models import (
    DEFAULT_FILE_EXTENSIONS,
    DatasetChunking,
    FileFormat,
    GeographicalExtent,
    TimeExtent,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.utils import (
    timestamp_or_datestring_to_datetime,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    TemporalParameters,
)

logger = logging.getLogger("copernicusmarine")


def get_file_extension(file_format: FileFormat) -> str:
    if file_format == "zarr":
        return ".zarr"
    if file_format == "csv":
        return ".csv"
    if file_format == "parquet":
        return ".parquet"
    else:
        return ".nc"


def get_filename(
    filename: Optional[str],
    dataset: xarray.Dataset,
    dataset_id: str,
    file_format: FileFormat,
    axis_coordinate_id_mapping: dict[str, str],
    geographical_parameters: GeographicalParameters,
) -> str:
    if filename:
        if pathlib.Path(filename).suffix in DEFAULT_FILE_EXTENSIONS:
            return filename
        else:
            return filename + get_file_extension(file_format)
    else:
        return _build_filename_from_dataset(
            dataset,
            dataset_id,
            file_format,
            axis_coordinate_id_mapping,
            geographical_parameters,
        )


def _build_filename_from_dataset(
    dataset: xarray.Dataset,
    dataset_id: str,
    file_format: FileFormat,
    axis_coordinate_id_mapping: dict[str, str],
    geographical_parameters: GeographicalParameters,
) -> str:
    dataset_variables = "-".join(
        [str(variable_name) for variable_name in dataset.data_vars]
    )
    variables = (
        "multi-vars"
        if (len(dataset_variables) > 15 and len(list(dataset.keys())) > 1)
        else dataset_variables
    )
    longitudes = None
    if "x" in axis_coordinate_id_mapping:
        if geographical_parameters.projection == "lonlat":
            longitudes = _format_longitudes(
                _get_min_coordinate(dataset, axis_coordinate_id_mapping["x"]),
                _get_max_coordinate(dataset, axis_coordinate_id_mapping["x"]),
            )
        if geographical_parameters.projection == "originalGrid":
            longitudes = _format_xy_axis(
                _get_min_coordinate(dataset, axis_coordinate_id_mapping["x"]),
                _get_max_coordinate(dataset, axis_coordinate_id_mapping["x"]),
                "x",
            )

    latitudes = None
    if "y" in axis_coordinate_id_mapping:
        if geographical_parameters.projection == "lonlat":
            latitudes = _format_latitudes(
                _get_min_coordinate(dataset, axis_coordinate_id_mapping["y"]),
                _get_max_coordinate(dataset, axis_coordinate_id_mapping["y"]),
            )
        if geographical_parameters.projection == "originalGrid":
            latitudes = _format_xy_axis(
                _get_min_coordinate(dataset, axis_coordinate_id_mapping["y"]),
                _get_max_coordinate(dataset, axis_coordinate_id_mapping["y"]),
                "y",
            )

    depths = None
    if "z" in axis_coordinate_id_mapping:
        depths = _format_depths(
            _get_min_coordinate(dataset, axis_coordinate_id_mapping["z"]),
            _get_max_coordinate(dataset, axis_coordinate_id_mapping["z"]),
        )

    datetimes = None
    if "t" in axis_coordinate_id_mapping:
        min_time_coordinate = _get_min_coordinate(
            dataset, axis_coordinate_id_mapping["t"]
        )
        max_time_coordinate = _get_max_coordinate(
            dataset, axis_coordinate_id_mapping["t"]
        )
        datetimes = _format_datetimes(
            (
                timestamp_or_datestring_to_datetime(min_time_coordinate)
                if min_time_coordinate is not None
                else None
            ),
            (
                timestamp_or_datestring_to_datetime(max_time_coordinate)
                if max_time_coordinate is not None
                else None
            ),
        )
    filename = "_".join(
        filter(
            None,
            [dataset_id, variables, longitudes, latitudes, depths, datetimes],
        )
    )
    filename = filename if len(filename) < 250 else filename[250:]

    return filename + get_file_extension(file_format)


def build_filename_from_request(
    request: SubsetRequest,
    variables: list[str],
    platform_ids: list[str],
    axis_coordinate_id_mapping: dict[str, str],
) -> str:
    """
    In the sparse dataset case we don't have the dataset to build the filename from.
    """  # noqa

    dataset_variables = "-".join(variables)
    dataset_variables = (
        "multi-vars" if len(variables) > 15 else dataset_variables
    )
    platform_ids_text = "-".join(platform_ids)
    platform_ids_text = (
        "multi-platforms" if len(platform_ids) > 15 else platform_ids_text
    )
    longitudes = None
    if "x" in axis_coordinate_id_mapping:
        if axis_coordinate_id_mapping["x"] == "longitude":
            longitudes = _format_longitudes(
                request.minimum_x, request.maximum_x
            )
        if axis_coordinate_id_mapping["x"] == "x":
            longitudes = _format_xy_axis(
                request.minimum_x,
                request.maximum_x,
                axis_coordinate_id_mapping["x"],
            )

    latitudes = None
    if "y" in axis_coordinate_id_mapping:
        if axis_coordinate_id_mapping["y"] == "latitude":
            latitudes = _format_latitudes(request.minimum_y, request.maximum_y)
        if axis_coordinate_id_mapping["y"] == "y":
            latitudes = _format_xy_axis(
                request.minimum_y,
                request.maximum_y,
                axis_coordinate_id_mapping["y"],
            )

    depths = None
    if "z" in axis_coordinate_id_mapping:
        if axis_coordinate_id_mapping["z"] == "depth":
            depths = _format_depths(
                request.minimum_depth, request.maximum_depth
            )

    datetimes = None
    if "t" in axis_coordinate_id_mapping:
        datetimes = _format_datetimes(
            request.start_datetime, request.end_datetime
        )
    filename = "_".join(
        filter(
            None,
            [
                request.dataset_id,
                dataset_variables,
                platform_ids_text,
                longitudes,
                latitudes,
                depths,
                datetimes,
            ],
        )
    )
    filename = filename if len(filename) < 250 else filename[250:]

    return filename + get_file_extension(request.file_format)


def get_coordinate_ids_from_parameters(
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
) -> list[str]:
    return [
        coordinate_id
        for coordinate_id in [
            geographical_parameters.x_axis_parameters.coordinate_id,
            geographical_parameters.y_axis_parameters.coordinate_id,
            temporal_parameters.coordinate_id,
            depth_parameters.coordinate_id,
        ]
        if coordinate_id
    ]


def _get_min_coordinate(dataset: xarray.Dataset, coordinate_id: str) -> Any:
    if coordinate_id in dataset.sizes:
        return min(dataset[coordinate_id].values)
    return None


def _get_max_coordinate(dataset: xarray.Dataset, coordinate_id: str) -> Any:
    if coordinate_id in dataset.sizes:
        return max(dataset[coordinate_id].values)
    return None


def _get_unit_coordinate(dataset: xarray.Dataset, coordinate_id: str) -> Any:
    if coordinate_id in dataset.sizes:
        return dataset[coordinate_id].attrs.get("units")
    return None


def _format_longitudes(
    minimum_longitude: Optional[float], maximum_longitude: Optional[float]
) -> str:
    if minimum_longitude is None or maximum_longitude is None:
        return ""
    else:
        if minimum_longitude == maximum_longitude:
            suffix = "W" if minimum_longitude < 0 else "E"
            longitude = f"{abs(minimum_longitude):.2f}{suffix}"
        else:
            minimum_suffix = "W" if minimum_longitude < 0 else "E"
            maximum_suffix = "W" if maximum_longitude < 0 else "E"
            longitude = (
                f"{abs(minimum_longitude):.2f}{minimum_suffix}-"
                f"{abs(maximum_longitude):.2f}{maximum_suffix}"
            )
        return longitude


def _format_latitudes(
    minimum_latitude: Optional[float], maximum_latitude: Optional[float]
) -> str:
    if minimum_latitude is None or maximum_latitude is None:
        return ""
    else:
        if minimum_latitude == maximum_latitude:
            suffix = "S" if minimum_latitude < 0 else "N"
            latitude = f"{abs(minimum_latitude):.2f}{suffix}"
        else:
            minimum_suffix = "S" if minimum_latitude < 0 else "N"
            maximum_suffix = "S" if maximum_latitude < 0 else "N"
            latitude = (
                f"{abs(minimum_latitude):.2f}{minimum_suffix}-"
                f"{abs(maximum_latitude):.2f}{maximum_suffix}"
            )
        return latitude


def _format_xy_axis(
    minimum_value: Optional[float],
    maximum_value: Optional[float],
    coordinate_id: str,
) -> str:
    if minimum_value is None or maximum_value is None:
        return ""
    else:
        if minimum_value == maximum_value:
            suffix = coordinate_id
            value = f"{(minimum_value):.2f}{suffix}"
        else:
            minimum_suffix = coordinate_id
            maximum_suffix = coordinate_id.upper()
            value = (
                f"{(minimum_value):.2f}{minimum_suffix}_"
                f"{(maximum_value):.2f}{maximum_suffix}"
            )
        return value


def _format_depths(
    minimum_depth: Optional[float], maximum_depth: Optional[float]
) -> str:
    if minimum_depth is None or maximum_depth is None:
        return ""
    else:
        if minimum_depth == maximum_depth:
            depth = f"{abs(minimum_depth):.2f}m"
        else:
            depth = f"{abs(minimum_depth):.2f}-{abs(maximum_depth):.2f}m"
        return depth


def _format_datetimes(
    minimum_datetime: Optional[datetime], maximum_datetime: Optional[datetime]
) -> str:
    if minimum_datetime is None or maximum_datetime is None:
        return ""
    else:
        if minimum_datetime == maximum_datetime:
            formatted_datetime = f"{minimum_datetime.strftime('%Y-%m-%d')}"
        else:
            formatted_datetime = (
                f"{minimum_datetime.strftime('%Y-%m-%d')}-"
                f"{maximum_datetime.strftime('%Y-%m-%d')}"
            )
        return formatted_datetime


def get_dataset_coordinates_extent(
    dataset: xarray.Dataset, axis_coordinate_id_mapping: dict[str, str]
) -> list[Union[GeographicalExtent, TimeExtent]]:
    coordinates_extent = []
    for coord_axis in ["x", "y", "t", "z"]:
        if coordinate_id := axis_coordinate_id_mapping.get(coord_axis):
            if coordinate_extent := _get_coordinate_extent(
                dataset, coordinate_id
            ):
                coordinates_extent.append(coordinate_extent)

    return coordinates_extent


def _get_coordinate_extent(
    dataset: xarray.Dataset,
    coordinate_id: str,
) -> Optional[Union[GeographicalExtent, TimeExtent]]:
    if coordinate_id in dataset.sizes:
        minimum = _get_min_coordinate(dataset, coordinate_id)
        maximum = _get_max_coordinate(dataset, coordinate_id)
        unit = _get_unit_coordinate(dataset, coordinate_id)
        if coordinate_id == "time":
            minimum = timestamp_or_datestring_to_datetime(minimum).isoformat()
            maximum = timestamp_or_datestring_to_datetime(maximum).isoformat()
            unit = "iso8601"
            return TimeExtent(
                minimum=minimum,
                maximum=maximum,
                unit=unit,
                coordinate_id=coordinate_id,
            )
        return GeographicalExtent(
            minimum=minimum,
            maximum=maximum,
            unit=unit,
            coordinate_id=coordinate_id,
        )
    return None


def get_approximation_size_final_result(
    dataset: xarray.Dataset, axis_coordinate_id_mapping: dict[str, str]
) -> Optional[float]:
    coordinates_size = 1
    variables_size = 0
    baseline_size = 0.013

    for variable in dataset.data_vars:
        variables_size += dataset[variable].encoding["dtype"].itemsize

    for coordinate_name in dataset.sizes:
        for coord_label in axis_coordinate_id_mapping:
            if coordinate_name == axis_coordinate_id_mapping[coord_label]:
                coordinates_size *= dataset[coordinate_name].size
    estimate_size = baseline_size + coordinates_size * variables_size / 1048e3

    return estimate_size


def get_approximation_size_data_downloaded(
    dataset: xarray.Dataset,
    dataset_chunking: DatasetChunking,
) -> Optional[float]:
    # TODO: Test it not sure how to, maybe ask if the chunk size is correct
    temp_dataset = dataset.copy()
    if "elevation" in dataset.sizes:
        temp_dataset["elevation"] = temp_dataset.elevation * (-1)
        temp_dataset = temp_dataset.rename({"elevation": "depth"})

    download_estimated_size = 0
    for variable_name in temp_dataset.data_vars:
        download_estimated_size += (
            dataset_chunking.get_number_values_variable(str(variable_name))
            * temp_dataset[list(temp_dataset.data_vars)[0]].dtype.itemsize
            / 1048e3
        )

    return download_estimated_size



class MinimumLongitudeGreaterThanMaximumLongitude(Exception):
    """
    Exception raised when the minimum longitude is greater than the maximum longitude.

    Please make sure the minimum longitude is less or equal than the maximum longitude.
    """

    pass


class VariableDoesNotExistInTheDataset(Exception):
    """
    Exception raised when the variable does not exist in the dataset.

    Please sure the variable exists in the dataset
    and/or that you use the standard name.
    """

    def __init__(self, variable):
        super().__init__(
            f"The variable '{variable}' is neither a variable or a standard name in"
            f" the dataset."
        )
        self.__setattr__(
            "custom_exception_message",
            f"The variable '{variable}' is neither a variable or a standard name in "
            f"the dataset.",
        )


class CoordinatesOutOfDatasetBounds(Exception):
    """
    Exception raised when the coordinates are out of the dataset bounds.

    Please make sure the coordinates are within the dataset bounds. If you are using the
    strict-inside mode, the coordinates must be within the dataset bounds.
    """

    def __init__(self, message: str):
        super().__init__(message)
        self.__setattr__("custom_exception_message", message)


class NetCDFCompressionNotAvailable(Exception):
    """
    Exception raised when the NetCDF compression is not available.

    Please make sure the NetCDF compression is available
    with the current Python libraries.
    """

    pass


class WrongDatetimeFormat(Exception):
    """
    Exception raised when the datetime format is wrong.

    Supported formats are:

    * the string "now"
    * all formats supported by dateutil Python library

    see `dateutil documentation page <https://dateutil.readthedocs.io/en/stable/parser.html>`_.
    """  # noqa

    pass


class FormatNotSupported(Exception):
    """
    Exception raised when the format is not supported for the command.
    Usually, it means that you are trying to subset a sparse dataset and
    for sparse datasets lazy loading is not available.
    Use :func:`copernicusmarine.read_dataframe` or :func:`copernicusmarine.subset` instead.

    Please try other commands or use datasets with the supported format.
    """  # noqa

    def __init__(
        self, format_type: str, command_type: str, recommended_command: str
    ):
        super().__init__(
            f"Lazy loading of format type '{format_type}' not "
            f"supported with command '{command_type}'. "
            f"You may want to look into '{recommended_command}' instead "
            f"to subset your data."
        )


class ServiceNotSupported(Exception):
    """
    Exception raised when the service type is not supported.

    Some services are not supported by the current implementation of the toolbox.
    """

    def __init__(self, service_type):
        super().__init__(f"Service type {service_type} not supported.")


class MutuallyExclusiveArguments(Exception):
    """
    Exception raised when mutually exclusive arguments are used together.

    Please make sure the arguments are not used together.
    """

    def __init__(self, arg1, arg2):
        super().__init__(
            f"Arguments '{arg1}' and '{arg2}' are mutually exclusive."
        )


class LonLatSubsetNotAvailableInOriginalGridDatasets(Exception):
    """
    Exception raised when using longitude and latitude subset on
    a original grid dataset.

    The options ``--maximum-longitude``, ``--minimum-longitude``,
    ``--maximum-latitude`` and ``--minimum-latitude`` cannot be
      used with 'originalGrid' dataset part.

    """

    def __init__(self):
        super().__init__(
            "You cannot specify longitude and latitude when using the 'originalGrid' "
            "dataset part. "
            "Try using ``--minimum-x``, ``--maximum-x``, ``--minimum-y`` and "
            "``--maximum-y``."
        )


class XYNotAvailableInNonOriginalGridDatasets(Exception):
    """
    Exception raised when using x and y subset on a non-original grid
    dataset.

    Please make sure the dataset part is 'originalGrid' when the options
    ``--minimum-x``, ``--maximum-x``, ``--minimum-y`` and ``--maximum-y``.
    """

    def __init__(self):
        super().__init__(
            "You cannot specify x and y when not using the 'originalGrid' dataset part."
            " Try using ``--maximum-longitude``, ``--minimum-longitude``, "
            "``--maximum-latitude`` and ``--minimum-latitude`` instead"
            " or make sure to specify the dataset_part."
        )


class DatasetUpdating(Exception):
    """
    Exception raised when the dataset is currently updating
    and the flag raise-if-updating is set to True.
    To avoid this exception, you can remove the flag from the query,
    request a subset of data before the updating start date or
    wait for the ARCO service update to be completed, which can
    take from 5 minutes to some hours.
    """

    def __init__(self, message: str):
        super().__init__(message)


class NotEnoughPlatformMetadata(Exception):
    """
    Exception raised when there is not enough platform metadata
    and user wants to perform subset on platform ids.

    Please contact the Copernicus Marine support team if needed.
    """

    def __init__(self):
        super().__init__(
            "Not enough platform metadata. "
            "Please make sure the platform metadata is available."
        )


class PlatformsSubsettingNotAvailable(Exception):
    """
    Exception raised when the subsetting on platforms is not available.


    Please make sure to not request platform ids for this dataset.
    """

    def __init__(self):
        super().__init__(
            "Subsetting on platforms is not available for this dataset. "
            "Please make sure not to request platform IDs for this dataset."
        )


class WrongPlatformID(Exception):
    """
    Exception raised when the requested platform ids are not in the list of platforms.

    Please make sure the platform id is in the list of platforms.
    Check the describe output and the "platformseries" service for more information.
    """  # noqa

    def __init__(self, platform_ids, platforms_metadata_url):
        super().__init__(
            f"None of the platform ids '{platform_ids}' are in the list of platforms."
            f" Please check the describe output and the platforms metadata at"
            f" '{platforms_metadata_url}' for more information."
        )


class ServiceDoesNotExistForCommand(Exception):
    """
    Exception raised when the service does not exist for the command.

    Please make sure the service exists for the command.
    """  # TODO: list available services per command

    def __init__(
        self,
        requested_service_name: str,
        command_name: str,
        available_services: list[str],
    ):
        super().__init__()
        self.__setattr__(
            "custom_exception_message",
            f"Service {requested_service_name} "
            f"does not exist for command {command_name}. "
            f"Possible service{'s' if len(available_services) > 1 else ''}: "
            f"{available_services}",
        )


class NoServiceAvailable(Exception):
    """
    Exception raised when no service is available for the dataset.

    We could not find a service for this dataset.
    Please make sure there is a service available for the dataset.
    """

    pass


class ServiceNotAvailable(Exception):
    """
    Exception raised when the service is not available for the dataset.

    Please make sure the service is available for the specific dataset.
    """

    pass



import logging
from typing import Optional

import xarray
import zarr

from copernicusmarine.core_functions.utils import parse_access_dataset_url

logger = logging.getLogger("copernicusmarine")


def open_zarr(
    dataset_url: str,
    copernicus_marine_username: Optional[str] = None,
    **kwargs,
) -> xarray.Dataset:
    (
        endpoint,
        bucket,
        root_path,
    ) = parse_access_dataset_url(dataset_url)
    if zarr.__version__.startswith("2"):
        from copernicusmarine.core_functions.custom_s3_store_zarr_v2 import (
            CustomS3StoreZarrV2,
        )

        logger.debug("Using custom store for Zarr Python library v2")
        store = CustomS3StoreZarrV2(
            endpoint=endpoint,
            bucket=bucket,
            root_path=root_path,
            copernicus_marine_username=copernicus_marine_username,
        )
        return xarray.open_zarr(store, **kwargs)
    else:
        from copernicusmarine.core_functions.custom_s3_store_zarr_v3 import (
            CustomS3StoreZarrV3,
        )

        logger.debug("Using custom store for Zarr Python library v3")
        store = CustomS3StoreZarrV3(
            endpoint=endpoint,
            bucket=bucket,
            root_path=root_path,
            copernicus_marine_username=copernicus_marine_username,
        )
        return xarray.open_zarr(store, **kwargs, zarr_format=2)



import copernicusmarine.logger as logger
from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineCatalogue,
    CopernicusMarineCoordinate,
    CopernicusMarineDataset,
    CopernicusMarinePart,
    CopernicusMarineProduct,
    CopernicusMarineService,
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
    CopernicusMarineVariable,
    CopernicusMarineVersion,
    CoperniusMarineServiceShortNames,
    DatasetNotFound,
    DatasetVersionNotFound,
    DatasetVersionPartNotFound,
    ProductNotFound,
    ServiceNotHandled,
)
from copernicusmarine.command_line_interface.utils import (
    OtherOptionsPassedWithCreateTemplate,
)
from copernicusmarine.core_functions.credentials_utils import (
    CouldNotConnectToAuthenticationSystem,
    CredentialsCannotBeNone,
    InvalidUsernameOrPassword,
)
from copernicusmarine.core_functions.exceptions import (
    CoordinatesOutOfDatasetBounds,
    DatasetUpdating,
    FormatNotSupported,
    MinimumLongitudeGreaterThanMaximumLongitude,
    NetCDFCompressionNotAvailable,
    NoServiceAvailable,
    NotEnoughPlatformMetadata,
    PlatformsSubsettingNotAvailable,
    ServiceDoesNotExistForCommand,
    ServiceNotAvailable,
    ServiceNotSupported,
    VariableDoesNotExistInTheDataset,
    WrongDatetimeFormat,
    WrongPlatformID,
)
from copernicusmarine.core_functions.fields_query_builder import (
    WrongFieldsError,
)
from copernicusmarine.core_functions.models import (
    FileGet,
    FileStatus,
    GeographicalExtent,
    ResponseGet,
    ResponseSubset,
    StatusCode,
    StatusMessage,
    TimeExtent,
)
from copernicusmarine.python_interface.describe import describe
from copernicusmarine.python_interface.get import get
from copernicusmarine.python_interface.login import login
from copernicusmarine.python_interface.open_dataset import (
    open_dataset,
)
from copernicusmarine.python_interface.read_dataframe import (
    read_dataframe,
)
from copernicusmarine.python_interface.subset import subset
from copernicusmarine.versioner import __version__

__all__ = [
    "CoordinatesOutOfDatasetBounds",
    "CopernicusMarineCatalogue",
    "CopernicusMarineCoordinate",
    "CopernicusMarineDataset",
    "CopernicusMarinePart",
    "CopernicusMarineProduct",
    "CopernicusMarineService",
    "CopernicusMarineServiceFormat",
    "CopernicusMarineServiceNames",
    "CopernicusMarineVariable",
    "CopernicusMarineVersion",
    "CoperniusMarineServiceShortNames",
    "CouldNotConnectToAuthenticationSystem",
    "CredentialsCannotBeNone",
    "DatasetNotFound",
    "DatasetVersionNotFound",
    "DatasetVersionPartNotFound",
    "FileGet",
    "FormatNotSupported",
    "GeographicalExtent",
    "InvalidUsernameOrPassword",
    "MinimumLongitudeGreaterThanMaximumLongitude",
    "NetCDFCompressionNotAvailable",
    "NoServiceAvailable",
    "NotEnoughPlatformMetadata",
    "OtherOptionsPassedWithCreateTemplate",
    "PlatformsSubsettingNotAvailable",
    "ProductNotFound",
    "ResponseGet",
    "ResponseSubset",
    "ServiceDoesNotExistForCommand",
    "ServiceNotAvailable",
    "ServiceNotHandled",
    "ServiceNotSupported",
    "StatusCode",
    "FileStatus",
    "StatusMessage",
    "TimeExtent",
    "VariableDoesNotExistInTheDataset",
    "WrongDatetimeFormat",
    "DatasetUpdating",
    "WrongFieldsError",
    "WrongPlatformID",
    "__version__",
    "describe",
    "get",
    "logger",
    "login",
    "open_dataset",
    "read_dataframe",
    "subset",
]



import pathlib
from datetime import datetime
from typing import List, Optional, Union

import pandas as pd

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    deprecated_python_option,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_VERTICAL_AXIS,
    CommandType,
    CoordinatesSelectionMethod,
    VerticalAxis,
)
from copernicusmarine.core_functions.request_structure import LoadRequest
from copernicusmarine.core_functions.utils import get_geographical_inputs
from copernicusmarine.download_functions.download_zarr import (
    read_dataframe_from_arco_series,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    TemporalParameters,
    XParameters,
    YParameters,
)
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.python_interface.load_utils import (
    load_data_object_from_load_request,
)
from copernicusmarine.python_interface.utils import homogenize_datetime


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def read_dataframe(
    dataset_id: str,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    minimum_x: Optional[float] = None,
    maximum_x: Optional[float] = None,
    minimum_y: Optional[float] = None,
    maximum_y: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS,  # noqa
    start_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    end_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    ),
    service: Optional[str] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    disable_progress_bar: bool = False,
    platform_ids: Optional[List[str]] = None,
) -> pd.DataFrame:
    """
    Immediately loads a Pandas DataFrame into memory from a specified dataset.

    Unlike 'lazy-loading,' the data is loaded as soon as this function is executed, which may be preferable when rapid access to the entire dataset is required, but may require careful memory management.


    Parameters
    ----------
    dataset_id : str
        The datasetID, required.
    dataset_version : str, optional
        Force the selection of a specific dataset version.
    dataset_part : str, optional
        Force the selection of a specific dataset part.
    username : str, optional
        The username for authentication.
    password : str, optional
        The password for authentication.
    variables : List[str], optional
        List of variable names to extract.
    minimum_longitude : float, optional
        Minimum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    maximum_longitude : float, optional
        Maximum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    minimum_latitude : float, optional
        Minimum latitude for the subset. Requires a float from -90 degrees to +90.
    maximum_latitude : float, optional
        Maximum latitude for the subset. Requires a float from -90 degrees to +90.
    minimum_x : float, optional
        Minimum x-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_x : float, optional
        Maximum x-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_y : float, optional
        Minimum y-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_y : float, optional
        Maximum y-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_depth : float, optional
        Minimum depth for the subset. Requires a positive float (or 0).
    maximum_depth : float, optional
        Maximum depth for the subset. Requires a positive float (or 0).
    vertical_axis : str, optional
        Consolidate the vertical dimension (the z-axis) as requested: depth with descending positive values, elevation with ascending positive values. Default is depth.
    start_datetime : Union[datetime, str], optional
        The start datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    end_datetime : Union[datetime, str], optional
        The end datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    coordinates_selection_method : str, optional
        If ``inside``, the selection retrieved will be inside the requested range. If ``strict-inside``, the selection retrieved will be inside the requested range, and an error will be raised if the values don't exist. If ``nearest``, the extremes closest to the requested values will be returned. If ``outside``, the extremes will be taken to contain all the requested interval. The methods ``inside``, ``nearest`` and ``outside`` will display a warning if the request is out of bounds.
    service : str, optional
        Force download through one of the available services using the service name among ['arco-geo-series', 'arco-time-series', 'omi-arco', 'static-arco', 'arco-platform-series'] or its short name among ['geoseries', 'timeseries', 'omi-arco', 'static-arco', 'platformseries'].
    credentials_file : Union[pathlib.Path, str], optional
        Path to a credentials file if not in its default directory (``$HOME/.copernicusmarine``). Accepts .copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini files.
    disable_progress_bar : bool, optional
        Flag to hide progress bar.
    platform_ids : List[str], optional
        List of platform IDs to extract. Only available for platform chunked datasets.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing the loaded Copernicus Marine data.
    """  # noqa

    start_datetime = homogenize_datetime(start_datetime)
    end_datetime = homogenize_datetime(end_datetime)
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )

    (
        minimum_x_axis,
        maximum_x_axis,
        minimum_y_axis,
        maximum_y_axis,
    ) = get_geographical_inputs(
        minimum_longitude,
        maximum_longitude,
        minimum_latitude,
        maximum_latitude,
        minimum_x,
        maximum_x,
        minimum_y,
        maximum_y,
        dataset_part,
    )

    geographicalparameters = GeographicalParameters(
        y_axis_parameters=YParameters(
            minimum_y=minimum_y_axis,
            maximum_y=maximum_y_axis,
            coordinate_id=(
                "y" if dataset_part == "originalGrid" else "latitude"
            ),
        ),
        x_axis_parameters=XParameters(
            minimum_x=minimum_x_axis,
            maximum_x=maximum_x_axis,
            coordinate_id=(
                "x" if dataset_part == "originalGrid" else "longitude"
            ),
        ),
        projection=(
            "originalGrid" if dataset_part == "originalGrid" else "lonlat"
        ),
    )

    load_request = LoadRequest(
        dataset_id=dataset_id,
        force_dataset_version=dataset_version,
        force_dataset_part=dataset_part,
        username=username,
        password=password,
        variables=variables,
        platform_ids=platform_ids,
        geographical_parameters=geographicalparameters,
        temporal_parameters=TemporalParameters(
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        ),
        coordinates_selection_method=coordinates_selection_method,
        depth_parameters=DepthParameters(
            minimum_depth=minimum_depth,
            maximum_depth=maximum_depth,
            vertical_axis=vertical_axis,
        ),
        force_service=service,
        credentials_file=credentials_file,
        disable_progress_bar=disable_progress_bar,
    )

    dataset = load_data_object_from_load_request(
        load_request,
        read_dataframe_from_arco_series,
        chunks_factor_size_limit=0,
        command_type=CommandType.READ_DATAFRAME,
    )
    return dataset



from typing import Optional

from copernicusmarine.catalogue_parser.models import CopernicusMarineCatalogue
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    deprecated_python_option,
)
from copernicusmarine.core_functions.describe import describe_function
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def describe(
    show_all_versions: bool = False,
    contains: list[str] = [],
    product_id: Optional[str] = None,
    dataset_id: Optional[str] = None,
    max_concurrent_requests: int = 15,
    disable_progress_bar: bool = False,
    staging: bool = False,
) -> CopernicusMarineCatalogue:
    """
    Retrieve and parse the metadata information from the Copernicus Marine catalogue.

    Parameters
    ----------
    show_all_versions : bool, optional
        Include dataset versions in output. By default, shows only the default version.
    contains : list[str], optional
        Filter catalogue output. Returns products with attributes matching a string token.
    product_id : str, optional
        Force the productID to be used for the describe command. Will not parse the whole catalogue, but only the product with the given productID.
    dataset_id : str, optional
        Force the datasetID to be used for the describe command. Will not parse the whole catalogue, but only the dataset with the given datasetID.
    max_concurrent_requests : int, optional
        Maximum number of concurrent requests (>=1). Default 15. The command uses a thread pool executor to manage concurrent requests.
    disable_progress_bar : bool, optional
        Flag to hide progress bar.

    Returns
    -------
    copernicusmarine.CopernicusMarineCatalogue
        An object containing the retrieved metadata information.

    """  # noqa

    if not isinstance(contains, list):
        raise ValueError("contains must be of list type")

    return describe_function(
        show_all_versions,
        contains,
        product_id,
        dataset_id,
        max_concurrent_requests,
        disable_progress_bar,
        staging=staging,
    )



import logging
import logging.config
import time

logging_configuration_dict = {
    "disable_existing_loggers": False,
    "formatters": {
        "blank": {"format": "%(message)s"},
        "simple": {
            "datefmt": "%Y-%m-%dT%H:%M:%SZ",
            "format": "%(levelname)s - %(asctime)s - %(message)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "simple",
            "level": "DEBUG",
            "stream": "ext://sys.stderr",
        },
        "console_blank": {
            "class": "logging.StreamHandler",
            "formatter": "blank",
            "level": "DEBUG",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "copernicusmarine_blank_logger": {
            "handlers": ["console_blank"],
            "level": "INFO",
            "propagate": False,
        },
        "copernicusmarine": {
            "handlers": ["console"],
            "level": "INFO",
        },
    },
    "version": 1,
}

logging.config.dictConfig(logging_configuration_dict)
logging.Formatter.converter = time.gmtime



import calendar
import concurrent.futures
import logging
import pathlib
import re
from datetime import datetime, timedelta, timezone
from typing import (
    Any,
    Callable,
    Iterator,
    Literal,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

import numpy
import xarray
from dateutil import parser
from dateutil.parser._parser import ParserError
from requests import PreparedRequest
from tqdm import tqdm

from copernicusmarine.core_functions.exceptions import (
    LonLatSubsetNotAvailableInOriginalGridDatasets,
    WrongDatetimeFormat,
    XYNotAvailableInNonOriginalGridDatasets,
)
from copernicusmarine.versioner import __version__ as copernicusmarine_version

logger = logging.getLogger("copernicusmarine")


def get_unique_filepath(
    filepath: pathlib.Path,
) -> pathlib.Path:
    parent = filepath.parent
    filename = filepath.stem
    extension = filepath.suffix
    counter = 1

    while filepath.exists():
        filepath = parent / (filename + "_(" + str(counter) + ")" + extension)
        counter += 1
    return filepath


_T = TypeVar("_T")


def next_or_raise_exception(
    iterator: Iterator[_T], exception_to_raise: Exception
) -> _T:
    try:
        return next(iterator)
    except Exception as exception:
        raise exception_to_raise from exception


def construct_url_with_query_params(url, query_params: dict) -> Optional[str]:
    req = PreparedRequest()
    req.prepare_url(url, query_params)
    return req.url


def construct_query_params_for_marine_data_store_monitoring(
    username: Optional[str] = None,
) -> dict:
    query_params = {
        "x-cop-client": "copernicus-marine-toolbox",
        "x-cop-client-version": copernicusmarine_version,
    }
    if username:
        query_params["x-cop-user"] = username
    return query_params


def datetime_parser(date: Union[str, numpy.datetime64]) -> datetime:
    if date == "now":
        return datetime.now(tz=timezone.utc)
    try:
        if isinstance(date, numpy.datetime64):
            date = str(date)
        parsed_datetime = parser.parse(
            date, default=datetime(1978, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc)
        )
        return parsed_datetime
    except ParserError:
        pass
    raise WrongDatetimeFormat(date)


def timestamp_parser(
    timestamp: Union[int, float], unit: Literal["s", "ms"] = "ms"
) -> datetime:
    """
    Convert a timestamp in milliseconds to a datetime object.
    The unit can be changed to seconds by passing "s".
    """
    delta = (
        timedelta(seconds=timestamp)
        if unit == "s"
        else timedelta(milliseconds=timestamp)
    )
    return datetime(1970, 1, 1, tzinfo=timezone.utc) + delta


def datetime_to_timestamp(
    date: datetime, unit: Literal["s", "ms"] = "ms"
) -> Union[int, float]:
    """
    Should be Windows compatible for datetime before 1970
    """
    return calendar.timegm(date.timetuple()) * (1000 if unit == "ms" else 1)


def timestamp_or_datestring_to_datetime(
    date: Union[str, int, float, numpy.datetime64],
) -> datetime:
    if isinstance(date, int) or isinstance(date, float):
        return timestamp_parser(date)
    else:
        return datetime_parser(date)


def datetime_to_isoformat(
    date: datetime,
) -> str:
    """
    Convert a datetime object to ISO 8601 format.
    For consistency we want to return a timezone-aware datetime.

    Example: 2023-11-25T00:00:00+00:00
    or 2023-11-25T00:00:00Z (we prefer the latter)
    """
    return date.isoformat().replace("+00:00", "Z")


def add_copernicusmarine_version_in_dataset_attributes(
    dataset: xarray.Dataset,
) -> xarray.Dataset:
    dataset.attrs["copernicusmarine_version"] = copernicusmarine_version
    return dataset


# From: https://stackoverflow.com/a/46144596/20983727
def run_concurrently(
    func: Callable[..., _T],
    function_arguments: Sequence[tuple[Any, ...]],
    max_concurrent_requests: int,
    tdqm_bar_configuration: dict = {},
) -> list[_T]:
    out = []
    with tqdm(
        total=len(function_arguments),
        **tdqm_bar_configuration,
    ) as pbar:
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=max_concurrent_requests
        ) as executor:
            future_to_url = (
                executor.submit(func, *function_argument)
                for function_argument in function_arguments
            )
            for future in concurrent.futures.as_completed(future_to_url):
                data = future.result()
                out.append(data)
                pbar.update(1)
    return out


# Example data_path
# https://s3.waw3-1.cloudferro.com/mdl-native-01/native/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m_202105
# https://s3.region.cloudferro.com/bucket/arco/product/dataset/geoChunked.zarr
# https://s3.region.cloudferro.com:443/bucket/arco/product/dataset/geoChunked.zarr
def parse_access_dataset_url(
    data_path: str, only_dataset_root_path: bool = False
) -> tuple[str, str, str]:

    match = re.search(
        r"^(http|https):\/\/([\w\-\.]+)(:[\d]+)?(\/.*)", data_path
    )
    if match:
        endpoint_url = match.group(1) + "://" + match.group(2)
        full_path = match.group(4)
        segments = full_path.split("/")
        bucket = segments[1]
        path = (
            "/".join(segments[2:])
            if not only_dataset_root_path
            else "/".join(segments[2:5]) + "/"
        )
        return endpoint_url, bucket, path
    else:
        raise Exception(f"Invalid data path: {data_path}")


def create_custom_query_function(username: Optional[str]) -> Callable:
    def _add_custom_query_param(params, context, **kwargs):
        """
        Add custom query params for MDS's Monitoring
        """
        params["url"] = construct_url_with_query_params(
            params["url"],
            construct_query_params_for_marine_data_store_monitoring(username),
        )

    return _add_custom_query_param


def get_geographical_inputs(
    minimum_longitude: Optional[float],
    maximum_longitude: Optional[float],
    minimum_latitude: Optional[float],
    maximum_latitude: Optional[float],
    minimum_x: Optional[float],
    maximum_x: Optional[float],
    minimum_y: Optional[float],
    maximum_y: Optional[float],
    dataset_part: Optional[str],
) -> tuple[Optional[float], Optional[float], Optional[float], Optional[float]]:
    """
    Returns the geographical selection of the user.

    Parameters
    ----------
    minimum_longitude : float
        Minimum longitude of the area of interest. For lat/lon datasets.
    maximum_longitude : float
        Maximum longitude of the area of interest. For lat/lon datasets.
    minimum_latitude : float
        Minimum latitude of the area of interest. For lat/lon datasets.
    maximum_latitude : float
        Maximum latitude of the area of interest. For lat/lon datasets.
    minimum_x : float
        Minimum x coordinate of the area of interest. For "originalGrid" datasets.
    maximum_x : float
        Maximum x coordinate of the area of interest. For "originalGrid" datasets.
    minimum_y : float
        Minimum y coordinate of the area of interest. For "originalGrid" datasets.
    maximum_y : float
        Maximum y coordinate of the area of interest. For "originalGrid" datasets.
    dataset_part : str
        The part of the dataset to be used. If "originalGrid", the x and y coordinates
        should be the inputs.

    Returns
    -------
    tuple[Optional[float], Optional[float], Optional[float], Optional[float]]
        The geographical selection of the user. (minimum_x_axis, maximum_x_axis, minimum_y_axis, maximum_y_axis).

    Raises
    ------

    LonLatSubsetNotAvailableInOriginalGridDatasets
        If the dataset is "originalGrid" and the user tries to use lat/lon coordinates.
    XYNotAvailableInNonOriginalGridDatasets
        If the dataset is not "originalGrid" and the user tries to use x/y coordinates.
    """  # noqa: E501
    if dataset_part == "originalGrid":
        if (
            minimum_longitude is not None
            or maximum_longitude is not None
            or minimum_latitude is not None
            or maximum_latitude is not None
        ):
            raise LonLatSubsetNotAvailableInOriginalGridDatasets
        else:
            return (
                minimum_x,
                maximum_x,
                minimum_y,
                maximum_y,
            )
    else:
        if (
            minimum_x is not None
            or maximum_x is not None
            or minimum_y is not None
            or maximum_y is not None
        ):
            raise XYNotAvailableInNonOriginalGridDatasets
        else:
            return (
                minimum_longitude,
                maximum_longitude,
                minimum_latitude,
                maximum_latitude,
            )



import pathlib
from datetime import datetime
from typing import List, Optional, Union

import pandas as pd

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    deprecated_python_option,
)
from copernicusmarine.core_functions.exceptions import (
    MutuallyExclusiveArguments,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_FILE_FORMAT,
    DEFAULT_VERTICAL_AXIS,
    CoordinatesSelectionMethod,
    FileFormat,
    ResponseSubset,
    VerticalAxis,
)
from copernicusmarine.core_functions.subset import subset_function
from copernicusmarine.core_functions.utils import get_geographical_inputs
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.python_interface.utils import homogenize_datetime


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def subset(
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS,  # noqa
    start_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    end_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    minimum_x: Optional[float] = None,
    maximum_x: Optional[float] = None,
    minimum_y: Optional[float] = None,
    maximum_y: Optional[float] = None,
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    ),
    output_filename: Optional[str] = None,
    file_format: FileFormat = DEFAULT_FILE_FORMAT,
    service: Optional[str] = None,
    request_file: Optional[Union[pathlib.Path, str]] = None,
    output_directory: Optional[Union[pathlib.Path, str]] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    motu_api_request: Optional[str] = None,
    overwrite: bool = False,
    skip_existing: bool = False,
    dry_run: bool = False,
    disable_progress_bar: bool = False,
    staging: bool = False,
    netcdf_compression_level: int = 0,
    netcdf3_compatible: bool = False,
    chunk_size_limit: int = -1,
    raise_if_updating: bool = False,
    platform_ids: Optional[List[str]] = None,
) -> ResponseSubset:
    """
    Extract a subset of data from a specified dataset using given parameters.

    The datasetID is required and can be found via the ``describe`` command.

    Parameters
    ----------
    dataset_id : str, optional
        The datasetID, required either as an argument or in the request_file option.
    dataset_version : str, optional
        Force the selection of a specific dataset version.
    dataset_part : str, optional
        Force the selection of a specific dataset part.
    username : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    password : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    variables : List[str], optional
        List of variable names to extract.
    minimum_longitude : float, optional
        Minimum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    maximum_longitude : float, optional
        Maximum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    minimum_latitude : float, optional
        Minimum latitude for the subset. Requires a float from -90 degrees to +90.
    maximum_latitude : float, optional
        Maximum latitude for the subset. Requires a float from -90 degrees to +90.
    minimum_x : float, optional
        Minimum x-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_x : float, optional
        Maximum x-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_y : float, optional
        Minimum y-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_y : float, optional
        Maximum y-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_depth : float, optional
        Minimum depth for the subset. Requires a positive float (or 0).
    maximum_depth : float, optional
        Maximum depth for the subset. Requires a positive float (or 0).
    vertical_axis : str, optional
        Consolidate the vertical dimension (the z-axis) as requested: depth with descending positive values, elevation with ascending positive values. Default is depth.
    start_datetime : Union[datetime, str], optional
        The start datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    end_datetime : Union[datetime, str], optional
        The end datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    coordinates_selection_method : str, optional
        If ``inside``, the selection retrieved will be inside the requested range. If ``strict-inside``, the selection retrieved will be inside the requested range, and an error will be raised if the values don't exist. If ``nearest``, the extremes closest to the requested values will be returned. If ``outside``, the extremes will be taken to contain all the requested interval. The methods ``inside``, ``nearest`` and ``outside`` will display a warning if the request is out of bounds.
    output_directory : Union[pathlib.Path, str], optional
        The destination folder for the downloaded files. Default is the current directory.
    credentials_file : Union[pathlib.Path, str], optional
        Path to a credentials file if not in its default directory (``$HOME/.copernicusmarine``). Accepts .copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini files.
    output_filename : str, optional
        Save the downloaded data with the given file name (under the output directory).
    file_format : str, optional
        Format of the downloaded dataset. Default to NetCDF '.nc'.
    overwrite : bool, optional
        If specified and if the file already exists on destination, then it will be overwritten. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').
    skip_existing : bool, optional
        If the files already exists where it would be downloaded, then the download is skipped for this file. By default, the toolbox creates a new file with a new index (eg 'filename_(1).nc').
    service : str, optional
        Force download through one of the available services using the service name among ['arco-geo-series', 'arco-time-series', 'omi-arco', 'static-arco', 'arco-platform-series'] or its short name among ['geoseries', 'timeseries', 'omi-arco', 'static-arco', 'platformseries'].
    request_file : Union[pathlib.Path, str], optional
        Option to pass a file containing the arguments. For more information please refer to the documentation or use option ``--create-template`` from the command line interface for an example template.
    motu_api_request : str, optional
        Option to pass a complete MOTU API request as a string. Caution, user has to replace double quotes " with single quotes ' in the request.
    dry_run : bool, optional
        If True, runs query without downloading data.
    netcdf_compression_level : int, optional
        Specify a compression level to apply on the NetCDF output file. A value of 0 means no compression, and 9 is the highest level of compression available.
    netcdf3_compatible : bool, optional
        Enable downloading the dataset in a netCDF3 compatible format.
    chunk_size_limit : int, default -1
        Limit the size of the chunks in the dask array. Default is set to -1 which behaves similarly to 'chunks=auto' from ``xarray``. Positive integer values and '-1' are accepted. This is an experimental feature.
    raise_if_updating : bool, default False
        If set, raises a :class:`copernicusmarine.DatasetUpdating` error if the dataset is being updated and the subset interval requested overpasses the updating start date of the dataset. Otherwise, a simple warning is displayed.
    platform_ids : List[str], optional
        List of platform IDs to extract. Only available for platform chunked datasets.

    Returns
    -------
    ResponseSubset
        A description of the downloaded data and its destination.

    """  # noqa
    if overwrite:
        if skip_existing:
            raise MutuallyExclusiveArguments("overwrite", "skip_existing")

    request_file = pathlib.Path(request_file) if request_file else None
    output_directory = (
        pathlib.Path(output_directory) if output_directory else None
    )
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )

    if variables is not None:
        _check_type(variables, list, "variables")
    if platform_ids is not None:
        _check_type(platform_ids, list, "platform_ids")

    start_datetime = homogenize_datetime(start_datetime)
    end_datetime = homogenize_datetime(end_datetime)

    (
        minimum_x_axis,
        maximum_x_axis,
        minimum_y_axis,
        maximum_y_axis,
    ) = get_geographical_inputs(
        minimum_longitude,
        maximum_longitude,
        minimum_latitude,
        maximum_latitude,
        minimum_x,
        maximum_x,
        minimum_y,
        maximum_y,
        dataset_part,
    )

    return subset_function(
        dataset_id,
        dataset_version,
        dataset_part,
        username,
        password,
        variables,
        minimum_x_axis,
        maximum_x_axis,
        minimum_y_axis,
        maximum_y_axis,
        minimum_depth,
        maximum_depth,
        vertical_axis,
        start_datetime,
        end_datetime,
        platform_ids,
        coordinates_selection_method,
        output_filename,
        file_format,
        service,
        request_file,
        output_directory,
        credentials_file,
        motu_api_request,
        overwrite,
        skip_existing,
        dry_run,
        disable_progress_bar,
        staging,
        netcdf_compression_level,
        netcdf3_compatible,
        chunk_size_limit,
        raise_if_updating,
    )


def _check_type(value, expected_type: type, name: str):
    if not isinstance(value, expected_type):
        raise TypeError(f"{name} must be of type {expected_type.__name__}")



import logging
from dataclasses import dataclass

import semver

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_CONFIG_URL,
)
from copernicusmarine.core_functions.sessions import JsonParserConnection
from copernicusmarine.versioner import __version__ as toolbox_version

logger = logging.getLogger("copernicusmarine")

MARINE_DATASTORE_CONFIG_URL_CDN = (
    "https://stac.marine.copernicus.eu/clients-config-v1"
)
MARINE_DATASTORE_CONFIG_URL_DIRECT = (
    COPERNICUSMARINE_CONFIG_URL
    or "https://s3.waw3-1.cloudferro.com/mdl-metadata/clientsConfigV1.json"
)
MARINE_DATASTORE_CONFIG_URL_STAGING = (
    "https://stac-dta.marine.copernicus.eu/clientsConfigV1.json"
)

MARINE_DATASTORE_SERVICES_MAPPING: dict[str, list[str]] = {
    "describe": ["mds", "mds/serverlessArco/meta"],
    "get": ["mds", "mds/serverlessNative", "mds/serverlessArco/meta"],
    "subset": ["mds", "mds/serverlessArco", "mds/serverlessArco/meta"],
}


@dataclass
class CatalogueConfig:
    stac_catalogue_url: str
    dataset_product_mapping_url: str
    root_metadata_url: str


@dataclass
class MarineDataStoreConfig:
    catalogues: list[CatalogueConfig]
    staging: bool


def get_config_and_check_version_describe(
    staging: bool,
) -> MarineDataStoreConfig:
    marine_datastore_config = _check_version("describe", staging)
    return marine_datastore_config


def get_config_and_check_version_get(staging: bool) -> MarineDataStoreConfig:
    marine_datastore_config = _check_version("get", staging)
    return marine_datastore_config


def get_config_and_check_version_subset(
    staging: bool,
) -> MarineDataStoreConfig:
    marine_datastore_config = _check_version("subset", staging)
    return marine_datastore_config


def _check_version(function_name: str, staging: bool) -> MarineDataStoreConfig:
    def create_error_message(required_version: str) -> str:
        return (
            f"Installed copernicusmarine version {toolbox_version} might "
            f"lead to unexpected results with the current backend services. "
            f"Minimum supported version is {required_version}. Please update. "
            f"You can find instructions to install the latest version here: "
            f"https://toolbox-docs.marine.copernicus.eu/"
        )

    (
        marine_datastore_versions,
        marine_datastore_config,
    ) = _get_required_versions_and_config(staging)
    for service in MARINE_DATASTORE_SERVICES_MAPPING[function_name]:
        required_version = marine_datastore_versions[service]
        try:
            if not semver.Version.parse(toolbox_version).match(
                required_version
            ):
                logger.debug(
                    f"Client version {toolbox_version} is not compatible with "
                    f"{service}. Service needs version {required_version}."
                )
                logger.error(create_error_message(required_version))
                return marine_datastore_config
        except ValueError:
            logger.warning(
                f"Using a pre-release or a non-official version "
                f"of the client. Client version: {toolbox_version}"
            )
            return marine_datastore_config
    return marine_datastore_config


def _get_required_versions_and_config(
    staging: bool,
) -> tuple[dict[str, str], MarineDataStoreConfig]:
    url_mds_versions = (
        MARINE_DATASTORE_CONFIG_URL_STAGING
        if staging
        else MARINE_DATASTORE_CONFIG_URL_DIRECT
    )
    logger.debug(f"Getting required versions from {url_mds_versions}")
    mds_config: dict = {}
    try:
        with JsonParserConnection(
            timeout=2, retries=0
        ) as connection_without_retries:
            mds_config = connection_without_retries.get_json_file(
                url_mds_versions,
            )
    except Exception as e:
        if staging:
            raise e
        else:
            with JsonParserConnection() as connection:
                mds_config = connection.get_json_file(
                    MARINE_DATASTORE_CONFIG_URL_CDN,
                )
    if not mds_config:
        raise ValueError(
            "Please check your internet connection. "
            "Also, that you have whitelisted the domains indicated in "
            "https://toolbox-docs.marine.copernicus.eu/"
        )

    return (
        mds_config["clientVersions"],
        MarineDataStoreConfig(
            catalogues=[
                CatalogueConfig(
                    stac_catalogue_url=catalogue["stac"],
                    dataset_product_mapping_url=catalogue["idMapping"],
                    root_metadata_url=(
                        root_metadata_url[:-1]
                        if (
                            root_metadata_url := catalogue["stacRoot"]
                        ).endswith("/")
                        else root_metadata_url
                    ),
                )
                for catalogue in mds_config["catalogues"]
            ],
            staging=staging,
        ),
    )



import json
import logging
import os
import pathlib
from typing import Optional

from copernicusmarine.core_functions.credentials_utils import (
    get_and_check_username_password,
)
from copernicusmarine.core_functions.marine_datastore_config import (
    MarineDataStoreConfig,
    get_config_and_check_version_get,
)
from copernicusmarine.core_functions.models import CommandType, ResponseGet
from copernicusmarine.core_functions.request_structure import (
    GetRequest,
    filter_to_regex,
    overload_regex_with_additionnal_filter,
)
from copernicusmarine.core_functions.services_utils import (
    RetrievalService,
    get_retrieval_service,
)
from copernicusmarine.core_functions.utils import get_unique_filepath
from copernicusmarine.download_functions.download_original_files import (
    download_original_files,
)

logger = logging.getLogger("copernicusmarine")


def get_function(
    dataset_id: Optional[str],
    force_dataset_version: Optional[str],
    force_dataset_part: Optional[str],
    username: Optional[str],
    password: Optional[str],
    no_directories: bool,
    output_directory: Optional[pathlib.Path],
    credentials_file: Optional[pathlib.Path],
    overwrite: bool,
    request_file: Optional[pathlib.Path],
    filter_option: Optional[str],
    regex: Optional[str],
    file_list_path: Optional[pathlib.Path],
    create_file_list: Optional[str],
    sync: bool,
    sync_delete: bool,
    index_parts: bool,
    skip_existing: bool,
    dry_run: bool,
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    staging: bool,
) -> ResponseGet:
    marine_datastore_config = get_config_and_check_version_get(staging)
    if staging:
        logger.warning(
            "Detecting staging flag for get command. "
            "Data will come from the staging environment."
        )

    get_request = GetRequest(dataset_id=dataset_id or "")
    if request_file:
        get_request.from_file(request_file)
    if not get_request.dataset_id:
        raise ValueError("Please provide a dataset id for a get request.")
    request_update_dict = {
        "force_dataset_version": force_dataset_version,
        "output_directory": output_directory,
    }
    get_request.update(request_update_dict)

    # Specific treatment for default values:
    # In order to not overload arguments with default values
    # TODO is this really useful?
    if force_dataset_version:
        get_request.force_dataset_version = force_dataset_version
    if force_dataset_part:
        get_request.force_dataset_part = force_dataset_part
    if no_directories:
        get_request.no_directories = no_directories

    if overwrite:
        get_request.overwrite = overwrite
    if skip_existing:
        get_request.skip_existing = skip_existing
    if no_directories:
        get_request.no_directories = no_directories

    if filter_option:
        get_request.regex = filter_to_regex(filter_option)
    if regex:
        get_request.regex = overload_regex_with_additionnal_filter(
            regex, get_request.regex
        )
    if sync or sync_delete:
        get_request.sync = True
        if not get_request.force_dataset_version:
            raise ValueError(
                "Sync requires to set a dataset version. "
                "Please use --dataset-version option."
            )
    if sync_delete:
        get_request.sync_delete = sync_delete
    if index_parts:
        get_request.index_parts = index_parts
        get_request.regex = overload_regex_with_additionnal_filter(
            filter_to_regex("*index_*"), get_request.regex
        )
    if create_file_list is not None:
        assert create_file_list.endswith(".txt") or create_file_list.endswith(
            ".csv"
        ), "Download file list must be a '.txt' or '.csv' file. "
        f"Got '{create_file_list}' instead."
    if file_list_path:
        direct_download_files = get_direct_download_files(file_list_path)
        if direct_download_files:
            get_request.direct_download = direct_download_files
    if create_file_list or dry_run:
        get_request.dry_run = True

    return _run_get_request(
        username=username,
        password=password,
        get_request=get_request,
        create_file_list=create_file_list,
        credentials_file=credentials_file,
        marine_datastore_config=marine_datastore_config,
        max_concurrent_requests=max_concurrent_requests,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )


def _run_get_request(
    username: Optional[str],
    password: Optional[str],
    get_request: GetRequest,
    create_file_list: Optional[str],
    credentials_file: Optional[pathlib.Path],
    marine_datastore_config: MarineDataStoreConfig,
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    staging: bool = False,
) -> ResponseGet:
    logger.debug("Checking username and password...")
    username, password = get_and_check_username_password(
        username, password, credentials_file
    )
    logger.debug("Checking dataset metadata...")

    retrieval_service: RetrievalService = get_retrieval_service(
        dataset_id=get_request.dataset_id,
        force_dataset_version_label=get_request.force_dataset_version,
        force_dataset_part_label=get_request.force_dataset_part,
        force_service_name_or_short_name=None,
        command_type=CommandType.GET,
        dataset_subset=None,
        marine_datastore_config=marine_datastore_config,
    )
    get_request.dataset_url = retrieval_service.uri
    downloaded_files = download_original_files(
        username,
        password,
        get_request,
        max_concurrent_requests,
        disable_progress_bar,
        create_file_list,
    )
    return downloaded_files


def create_get_template() -> None:
    filename = pathlib.Path("get_template.json")
    if filename.exists():
        get_unique_filepath(
            filepath=pathlib.Path("get_template.json"),
        )
    with open(filename, "w") as output_file:
        json.dump(
            {
                "dataset_id": "cmems_mod_ibi_phy_my_0.083deg-3D_P1Y-m",
                "dataset_version": None,
                "dataset_part": None,
                "username": None,
                "password": None,
                "no_directories": False,
                "filter": "*01yav_200[0-2]*",
                "regex": None,
                "output_directory": "copernicusmarine_data",
                "file_list": None,
                "sync": False,
                "sync_delete": False,
                "index_parts": False,
                "disable_progress_bar": False,
                "overwrite": False,
                "log_level": "INFO",
                "dry_run": False,
            },
            output_file,
            indent=4,
        )
    logger.info(f"Template created at: {filename}")


def get_direct_download_files(
    file_list_path: pathlib.Path,
) -> list[str]:
    if not os.path.exists(file_list_path):
        raise FileNotFoundError(
            f"File {file_list_path} does not exist."
            " Please provide a valid path to a '.txt' file."
        )
    with open(file_list_path) as f:
        direct_download_files = [line.strip() for line in f.readlines()]
    return direct_download_files



from datetime import datetime
from typing import Optional, Union

import pandas as pd
from dateutil.tz import UTC

from copernicusmarine.core_functions.utils import datetime_parser


def homogenize_datetime(
    input_datetime: Optional[Union[datetime, pd.Timestamp, str]]
) -> Optional[datetime]:
    if input_datetime is None:
        return None
    if isinstance(input_datetime, str):
        return datetime_parser(input_datetime)
    if isinstance(input_datetime, pd.Timestamp):
        return input_datetime.to_pydatetime().astimezone(UTC)
    return input_datetime.astimezone(UTC)



import re
from dataclasses import dataclass
from enum import Enum
from typing import Literal, Optional, Type, TypeVar, Union

import pystac
from pydantic import BaseModel, ConfigDict

from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_debug,
)
from copernicusmarine.core_functions.utils import (
    datetime_parser,
    next_or_raise_exception,
)

VERSION_DEFAULT = "default"
PART_DEFAULT = "default"


class CopernicusMarineServiceNames(str, Enum):
    """
    Services parsed by the Copernicus Marine Toolbox.
    """

    GEOSERIES = "arco-geo-series"
    TIMESERIES = "arco-time-series"
    PLATFORMSERIES = "arco-platform-series"
    FILES = "original-files"
    WMTS = "wmts"
    OMI_ARCO = "omi-arco"
    STATIC_ARCO = "static-arco"


class CoperniusMarineServiceShortNames(str, Enum):
    """
    Short names or the services parsed by the Copernicus Marine Toolbox.
    Also accepted when a service is requested.
    """

    GEOSERIES = "geoseries"
    TIMESERIES = "timeseries"
    PLATFORMSERIES = "platformseries"
    FILES = "files"
    WMTS = "wmts"
    OMI_ARCO = "omi-arco"
    STATIC_ARCO = "static-arco"


def short_name_from_service_name(
    service_name: CopernicusMarineServiceNames,
) -> CoperniusMarineServiceShortNames:
    mapping = {
        CopernicusMarineServiceNames.GEOSERIES: CoperniusMarineServiceShortNames.GEOSERIES,  # noqa
        CopernicusMarineServiceNames.TIMESERIES: CoperniusMarineServiceShortNames.TIMESERIES,  # noqa
        CopernicusMarineServiceNames.PLATFORMSERIES: CoperniusMarineServiceShortNames.PLATFORMSERIES,  # noqa
        CopernicusMarineServiceNames.FILES: CoperniusMarineServiceShortNames.FILES,  # noqa
        CopernicusMarineServiceNames.WMTS: CoperniusMarineServiceShortNames.WMTS,  # noqa
        CopernicusMarineServiceNames.OMI_ARCO: CoperniusMarineServiceShortNames.OMI_ARCO,  # noqa
        CopernicusMarineServiceNames.STATIC_ARCO: CoperniusMarineServiceShortNames.STATIC_ARCO,  # noqa
    }
    return mapping[service_name]


def _service_type_from_web_api_string(
    name: str,
) -> CopernicusMarineServiceNames:
    class WebApi(Enum):
        GEOSERIES = "timeChunked"
        TIMESERIES = "geoChunked"
        PLATFORMSERIES = "platformChunked"
        FILES = "native"
        WMTS = "wmts"
        OMI_ARCO = "omi"
        STATIC_ARCO = "static"

    web_api_mapping: dict[WebApi, CopernicusMarineServiceNames] = {
        WebApi.GEOSERIES: CopernicusMarineServiceNames.GEOSERIES,
        WebApi.TIMESERIES: CopernicusMarineServiceNames.TIMESERIES,
        WebApi.PLATFORMSERIES: CopernicusMarineServiceNames.PLATFORMSERIES,
        WebApi.FILES: CopernicusMarineServiceNames.FILES,
        WebApi.WMTS: CopernicusMarineServiceNames.WMTS,
        WebApi.OMI_ARCO: CopernicusMarineServiceNames.OMI_ARCO,
        WebApi.STATIC_ARCO: CopernicusMarineServiceNames.STATIC_ARCO,
    }

    return next_or_raise_exception(
        (
            service_type
            for service_web_api, service_type in web_api_mapping.items()
            if service_web_api.value == name
        ),
        ServiceNotHandled(name),
    )


class ServiceNotHandled(Exception):
    """
    Exception raised when the dataset does not support the service type requested.

    Please verify that the requested service type can be found in
    the result of the :func:`~copernicusmarine.describe` command
    for this specific dataset, version and part.
    """

    pass


# service formats
class CopernicusMarineServiceFormat(str, Enum):
    """
    Format of the data for a service.
    For example, "arco-geo-series" and "arco-time-series" can be "zarr" or "sqlite".
    """

    ZARR = "zarr"
    SQLITE = "sqlite"


Coordinate = TypeVar("Coordinate", bound="CopernicusMarineCoordinate")


class CopernicusMarineCoordinate(BaseModel):
    """
    Coordinate for a variable.
    """

    #: Coordinate id.
    coordinate_id: str
    #: Coordinate units.
    coordinate_unit: str
    #: Minimum value of the coordinate.
    minimum_value: Optional[Union[float, str]]
    #: Maximum value of the coordinate.
    maximum_value: Optional[Union[float, str]]
    #: Step of the coordinate.
    step: Optional[float]
    #: Values of the coordinate.
    values: Optional[list[Union[float, int, str]]]
    #: Chunking length of the coordinate.
    chunking_length: Optional[Union[float, int]]
    #: Chunk type of the coordinate.
    chunk_type: Optional[str]
    #: Chunk reference coordinate of the coordinate.
    chunk_reference_coordinate: Optional[Union[float, int]]
    #: Chunk geometric factor of the coordinate.
    chunk_geometric_factor: Optional[Union[float, int]]
    #: Axis of the coordinate
    axis: Literal["x", "y", "z", "t"]

    @classmethod
    def from_metadata_item(
        cls: Type[Coordinate],
        variable_id: str,
        dimension: str,
        dimension_metadata: dict,
        arco_data_metadata_producer_valid_start_date: Optional[str],
        arco_data_metadata_producer_valid_start_index: Optional[int],
        cube_dimensions: dict,
    ) -> Coordinate:
        coordinates_info = dimension_metadata.get("coords", {})
        minimum_value = None
        coordinate_values = None
        if dimension == "time":
            if (
                arco_data_metadata_producer_valid_start_date
            ) and coordinates_info.get("min"):
                minimum_value = (
                    CopernicusMarineCoordinate._format_admp_valid_start_date(
                        arco_data_metadata_producer_valid_start_date,
                        to_timestamp=isinstance(
                            coordinates_info.get("min"), int
                        ),
                    )
                )
            elif (
                arco_data_metadata_producer_valid_start_index
                and coordinates_info.get("values")
            ):
                coordinate_values = coordinates_info.get("values")[
                    arco_data_metadata_producer_valid_start_index:
                ]
        chunking_length = dimension_metadata.get("chunkLen")
        axis = cube_dimensions[dimension].get("axis", "t")
        if isinstance(chunking_length, dict):
            chunking_length = chunking_length.get(variable_id)

        coordinate = cls(
            coordinate_id=dimension,
            coordinate_unit=dimension_metadata.get("units") or "",
            minimum_value=minimum_value or coordinates_info.get("min"),  # type: ignore
            maximum_value=coordinates_info.get("max"),
            step=coordinates_info.get("step"),
            values=coordinate_values or coordinates_info.get("values"),
            chunking_length=chunking_length,
            chunk_type=dimension_metadata.get("chunkType"),
            chunk_reference_coordinate=dimension_metadata.get("chunkRefCoord"),
            chunk_geometric_factor=dimension_metadata.get(
                "chunkGeometricFactor", {}
            ).get(variable_id),
            axis=axis,
        )
        if dimension == "elevation":
            coordinate._convert_elevation_to_depth()
        return coordinate

    @staticmethod
    def _format_admp_valid_start_date(
        arco_data_metadata_producer_valid_start_date: str,
        to_timestamp: bool = False,
    ) -> Union[str, int]:
        if to_timestamp:
            return int(
                datetime_parser(
                    arco_data_metadata_producer_valid_start_date
                ).timestamp()
                * 1000
            )
        return arco_data_metadata_producer_valid_start_date

    def _convert_elevation_to_depth(self):
        self.coordinate_id = "depth"
        minimum_elevation = self.minimum_value
        maximum_elevation = self.maximum_value
        if minimum_elevation is not None and isinstance(
            minimum_elevation, (int, float)
        ):
            self.maximum_value = -minimum_elevation
        else:
            self.maximum_value = None
        if maximum_elevation is not None and isinstance(
            maximum_elevation, (int, float)
        ):
            self.minimum_value = -maximum_elevation
        else:
            self.minimum_value = None
        if self.values is not None:
            self.values = [-value for value in self.values]  # type: ignore


Variable = TypeVar("Variable", bound="CopernicusMarineVariable")


class CopernicusMarineVariable(BaseModel):
    """
    Variable of the dataset.
    Contains the variable metadata and a list of coordinates.
    """

    #: Short name of the variable.
    short_name: str
    #: Standard name of the variable.
    standard_name: Optional[str]
    #: Units of the variable.
    units: Optional[str]
    #: Bounding box of the variable.
    bbox: Optional[list[float]]
    #: List of coordinates of the variable.
    coordinates: list[CopernicusMarineCoordinate]

    @classmethod
    def from_metadata_item(
        cls: Type[Variable],
        metadata_item: pystac.Item,
        asset: pystac.Asset,
        variable_id: str,
        bbox: Optional[list[float]],
    ) -> Variable:
        cube_variables = metadata_item.properties["cube:variables"]
        cube_variable = cube_variables[variable_id]
        cube_dimensions = metadata_item.properties["cube:dimensions"]
        extra_fields_asset = asset.extra_fields
        dimensions = extra_fields_asset.get("viewDims") or {}
        return cls(
            short_name=variable_id,
            standard_name=cube_variable["standardName"],
            units=cube_variable.get("unit") or "",
            bbox=bbox,
            coordinates=[
                CopernicusMarineCoordinate.from_metadata_item(
                    variable_id,
                    dimension,
                    dimension_metadata,
                    metadata_item.properties.get("admp_valid_start_date"),
                    metadata_item.properties.get("admp_valid_start_index"),
                    cube_dimensions,
                )
                for dimension, dimension_metadata in dimensions.items()
                if dimension in cube_variable["dimensions"]
            ],
        )


Service = TypeVar("Service", bound="CopernicusMarineService")


class CopernicusMarineService(BaseModel):
    """
    Service available for a dataset.
    Contains the service metadata and a list of variables.
    For original files service, there are no variables.
    """

    model_config = ConfigDict(use_enum_values=True)

    #: Service name.
    service_name: CopernicusMarineServiceNames

    #: Service short name.
    service_short_name: Optional[CoperniusMarineServiceShortNames]

    #: Service format: format of the service
    #: (eg:"arco-geo-series" can be "zarr", "sqlite").
    service_format: Optional[CopernicusMarineServiceFormat]
    #: Service uri: uri of the service.
    uri: str
    #: List of variables of the service.
    variables: list[CopernicusMarineVariable]
    #: A link to information about available platforms.
    #: Only for arco-platform-series service.
    platforms_metadata: Optional[str]

    @classmethod
    def from_metadata_item(
        cls: Type[Service],
        metadata_item: pystac.Item,
        service_name: str,
        asset: pystac.Asset,
    ) -> Optional[Service]:
        try:
            service_uri = asset.get_absolute_href()
            if not service_uri:
                raise ServiceNotHandled(service_name)
            service_name_parsed = _service_type_from_web_api_string(
                service_name
            )
            service_short_name = short_name_from_service_name(
                service_name_parsed
            )
            service_format = None
            admp_in_preparation = metadata_item.properties.get(
                "admp_in_preparation"
            )
            if asset.media_type and "zarr" in asset.media_type:
                service_format = CopernicusMarineServiceFormat.ZARR
            elif asset.media_type and "sqlite3" in asset.media_type:
                service_format = CopernicusMarineServiceFormat.SQLITE

            if not service_uri.endswith("/"):
                if admp_in_preparation and (
                    service_name_parsed
                    == CopernicusMarineServiceNames.GEOSERIES
                    or service_name_parsed
                    == CopernicusMarineServiceNames.TIMESERIES
                ):
                    return None
                else:
                    platforms_metadata = None
                    if (
                        service_name_parsed
                        == CopernicusMarineServiceNames.PLATFORMSERIES
                    ):
                        platforms_asset = metadata_item.get_assets().get(
                            "platforms"
                        )
                        if platforms_asset is not None:
                            platforms_metadata = platforms_asset.href

                    bbox = metadata_item.bbox
                    return cls(
                        service_name=service_name_parsed,
                        service_short_name=service_short_name,
                        uri=service_uri,
                        variables=[
                            CopernicusMarineVariable.from_metadata_item(
                                metadata_item, asset, var_cube["id"], bbox
                            )
                            for var_cube in metadata_item.properties[
                                "cube:variables"
                            ].values()
                        ],
                        service_format=service_format,
                        platforms_metadata=platforms_metadata,
                    )
            return None
        except ServiceNotHandled as service_not_handled:
            log_exception_debug(service_not_handled)
            return None

    def get_axis_coordinate_id_mapping(
        self,
    ) -> dict[str, str]:
        axis_coordinate_id_mapping: dict[str, str] = {}
        for variable in self.variables:
            for coordinate in variable.coordinates:
                if len(axis_coordinate_id_mapping) == 4:
                    return axis_coordinate_id_mapping
                axis_coordinate_id_mapping[
                    coordinate.axis
                ] = coordinate.coordinate_id

        return axis_coordinate_id_mapping


VersionPart = TypeVar("VersionPart", bound="CopernicusMarinePart")


class CopernicusMarinePart(BaseModel):
    """
    Part of a dataset. Datasets can have multiple parts.
    Each part contains a distinct list of services and distinct data.
    """

    #: Name of the part.
    name: str
    #: List of services available for the part.
    services: list[CopernicusMarineService]
    #: Date when the part will be retired.
    retired_date: Optional[str]
    #: Date when the part will be/was released.
    released_date: Optional[str]
    #: Date (of the data) starting from which the data is currently being updated.
    #: If set, the data after this date may not be up to date.
    #: Only applies to ARCO series
    #: and not to the original files.
    arco_updating_start_date: Optional[str]
    #: Date when the arco series of the part were last updated.
    #: Only applies to ARCO series
    #: and not to the original files.
    arco_updated_date: Optional[str]
    #: TODO: ask if this should be hidden
    # = Field(..., exclude=True)
    # if yes: needs to modify the query builder
    url_metadata: str

    @classmethod
    def from_metadata_item(
        cls: Type[VersionPart],
        metadata_item: pystac.Item,
        part_name: str,
        url_metadata: str,
    ) -> Optional[VersionPart]:
        retired_date = metadata_item.properties.get("admp_retired_date")
        released_date = metadata_item.properties.get("admp_released_date")
        arco_updated_date = metadata_item.properties.get("admp_updated_data")
        arco_updating_start_date = metadata_item.properties.get(
            "admp_updating_start_date"
        )
        if retired_date and datetime_parser(retired_date) < datetime_parser(
            "now"
        ):
            return None
        services = [
            service
            for metadata_service_name, asset in metadata_item.get_assets().items()
            if (
                service := CopernicusMarineService.from_metadata_item(
                    metadata_item,
                    metadata_service_name,
                    asset,
                )
            )
        ]
        if not services:
            return None
        services = services
        return cls(
            name=part_name,
            services=services,
            retired_date=retired_date,
            released_date=released_date,
            arco_updated_date=arco_updated_date,
            arco_updating_start_date=arco_updating_start_date,
            url_metadata=url_metadata,
        )

    def get_service_by_service_name(
        self, service_name: CopernicusMarineServiceNames
    ) -> CopernicusMarineService:
        return next(
            service
            for service in self.services
            if service.service_name == service_name
        )

    def get_coordinates(
        self,
    ) -> dict[
        str,
        tuple[
            CopernicusMarineCoordinate,
            list[str],
            list[CopernicusMarineServiceNames],
        ],
    ]:
        """
        Get the coordinates of the part as a dict.
        The dict has the coordinate IDs as keys and the values are tuples of:

        - the coordinate
        - variable_ids: list of variables the coordinate is associated with
        - service_names: list of service names the coordinate is associated with

        Returns
        -------
        dict
            The coordinates of the part and the associated variables and services
        """
        coordinates = {}
        variables = []
        services = []
        for service in self.services:
            services.append(service.service_name)
            for variable in service.variables:
                variables.append(variable.short_name)
                for coordinate in variable.coordinates:
                    coordinate_id = coordinate.coordinate_id
                    if coordinate_id not in coordinates:
                        coordinates[coordinate_id] = (
                            coordinate,
                            [variable.short_name],
                            [service.service_name],
                        )
                    else:
                        coordinates[coordinate_id][1].append(
                            variable.short_name
                        )
                        coordinates[coordinate_id][2].append(
                            service.service_name
                        )
        return coordinates


class CopernicusMarineVersion(BaseModel):
    """
    Version of a dataset. Datasets can have multiple versions.
    Usually around data releases.
    """

    #: Label of the version (eg: "latest", "202101").
    label: str
    #: List of parts of the version.
    parts: list[CopernicusMarinePart]

    def get_part(self, force_part: Optional[str]) -> CopernicusMarinePart:
        wanted_part = force_part or PART_DEFAULT
        for part in self.parts:
            if part.name == wanted_part:
                return part
            elif not force_part:
                return part
        raise DatasetVersionPartNotFound(self)

    def sort_parts(self) -> tuple[Optional[str], Optional[str]]:
        not_released_parts = {
            part.name
            for part in self.parts
            if part.released_date
            and datetime_parser(part.released_date) > datetime_parser("now")
        }
        will_be_retired_parts = {
            part.name: datetime_parser(part.retired_date).timestamp()
            for part in self.parts
            if part.retired_date
        }
        max_retired_timestamp = 0.0
        if will_be_retired_parts:
            max_retired_timestamp = max(will_be_retired_parts.values()) + 1
        self.parts = sorted(
            self.parts,
            key=lambda x: (
                x.name in not_released_parts,
                max_retired_timestamp
                - will_be_retired_parts.get(x.name, max_retired_timestamp),
                -(x.name == PART_DEFAULT),
                -(x.name == "latest"),  # for INSITU datasets
                -(x.name == "bathy"),  # for STATIC datasets
                x.name,
            ),
        )
        return self.parts[0].released_date, self.parts[0].retired_date


# For internal use only
@dataclass
class DatasetItem:
    """
    Intermediate class for the dataset item.
    Used to parse the dataset item from the catalogue.
    """

    item_id: str
    parsed_id: str
    parsed_part: str
    parsed_version: str
    url: str
    stac_item: pystac.Item
    stac_json: dict
    product_doi: Optional[str]


class CopernicusMarineDataset(BaseModel):
    """
    Dataset of a product.
    Contains the dataset metadata and a list of versions.
    """

    #: The datasetID.
    dataset_id: str
    #: The dataset name.
    dataset_name: str
    #: Digital object identifier or doi from
    #: the product the dataset belongs to.
    digital_object_identifier: Optional[str]
    #: List of versions of the dataset.
    versions: list[CopernicusMarineVersion]

    def get_version(
        self, force_version: Optional[str]
    ) -> CopernicusMarineVersion:
        wanted_version = force_version or VERSION_DEFAULT
        for version in self.versions:
            if version.label == wanted_version:
                return version
            elif not force_version:
                return version
        raise DatasetVersionNotFound(self)

    def sort_versions(self) -> None:
        not_released_versions: set[str] = set()
        retired_dates = {}
        for version in self.versions:
            released_date, retired_date = version.sort_parts()
            if released_date and datetime_parser(
                released_date
            ) > datetime_parser("now"):
                not_released_versions.add(version.label)
            if retired_date:
                retired_dates[version.label] = retired_date

        self.versions = sorted(
            self.versions,
            key=lambda x: (
                -(x.label in not_released_versions),
                retired_dates.get(x.label, "9999-12-31"),
                -(x.label == VERSION_DEFAULT),
                x.label,
            ),
            reverse=True,
        )

    def parse_dataset_metadata_items(
        self,
        dataset_items: list[DatasetItem],
    ) -> None:
        all_versions = set()
        for dataset_item in dataset_items:
            dataset_version = dataset_item.parsed_version
            part = CopernicusMarinePart.from_metadata_item(
                dataset_item.stac_item,
                dataset_item.parsed_part,
                dataset_item.url,
            )
            if not part:
                continue
            if dataset_version in all_versions:
                for version in self.versions:
                    if version.label == dataset_version:
                        version.parts.append(part)
                        break
            else:
                all_versions.add(dataset_version)
                version = CopernicusMarineVersion(
                    label=dataset_version, parts=[part]
                )
                self.versions.append(version)


class CopernicusMarineProduct(BaseModel):
    """
    Product of the catalogue.
    Contains the product metadata and a list of datasets.
    """

    #: Title of the product.
    title: str
    #: ProductID.
    product_id: str
    #: Thumbnail url of the product.
    thumbnail_url: str
    #: Description of the product.
    description: Optional[str]
    #: Digital object identifier or doi of the product.
    digital_object_identifier: Optional[str]
    #: Sources of the product.
    sources: list[str]
    #: Processing level of the product.
    processing_level: Optional[str]
    #: Production center of the product.
    production_center: str
    #: Keywords of the product.
    keywords: Optional[list[str]]
    #: List of datasets of the product.
    datasets: list[CopernicusMarineDataset]


class CopernicusMarineCatalogue(BaseModel):
    """
    Catalogue of the Copernicus Marine service.
    You can find here the products of the catalogue and their metadata as the response of the describe command/function.
    """  # noqa

    #: List of products in the catalogue.
    products: list[CopernicusMarineProduct]

    def filter_only_official_versions_and_parts(self):
        products_to_remove = []
        for product in self.products:
            datasets_to_remove = []
            for dataset in product.datasets:
                latest_version = dataset.versions[0]
                parts_to_remove = []
                for part in latest_version.parts:
                    if part.released_date and datetime_parser(
                        part.released_date
                    ) > datetime_parser("now"):
                        parts_to_remove.append(part)
                for part_to_remove in parts_to_remove:
                    latest_version.parts.remove(part_to_remove)
                if not latest_version.parts:
                    datasets_to_remove.append(dataset)
                else:
                    dataset.versions = [latest_version]
            for dataset_to_remove in datasets_to_remove:
                product.datasets.remove(dataset_to_remove)
            if not product.datasets:
                products_to_remove.append(product)
        for product_to_remove in products_to_remove:
            self.products.remove(product_to_remove)


# Errors
class DatasetVersionPartNotFound(Exception):
    """
    Exception raised when the asked part of the version of the dataset cannot be found.

    Please verify that the requested part can be found in
    the result of the :func:`~copernicusmarine.describe` command
    for this specific dataset version and datasetID.
    If yes, please contact the User Support, (widget chat on
    `Copernicus Marine website <https://help.marine.copernicus.eu/en/>`_).
    """

    def __init__(self, version: CopernicusMarineVersion):
        message = f"No part found for version {version.label}"
        super().__init__(message)


class DatasetVersionNotFound(Exception):
    """
    Exception raised when the asked version of the dataset cannot be found.

    Please verify that the requested version can be found in
    the result of the :func:`~copernicusmarine.describe` command
    for this specific dataset.
    If yes, please contact the User Support, (widget chat on
    `Copernicus Marine website <https://help.marine.copernicus.eu/en/>`_).
    """

    def __init__(self, dataset: CopernicusMarineDataset):
        message = f"No version found for dataset {dataset.dataset_id}"
        super().__init__(message)


class DatasetNotFound(Exception):
    """
    Exception raised when the dataset is not found in the catalogue.

    Possible reasons:

    - The datasetID is incorrect and not present in the catalogue.
    - The dataset has been retired.

    Please verify that the datasetID can be found in
    the result of the :func:`~copernicusmarine.describe` command.
    If yes, please contact the User Support, (widget chat on
    `Copernicus Marine website <https://help.marine.copernicus.eu/en/>`_).
    """

    def __init__(self, dataset_id: str):
        message = (
            f"{dataset_id} "
            f"Please check that the dataset exists and "
            f"the input datasetID is correct."
        )
        super().__init__(message)


class DatasetIsNotPartOfTheProduct(Exception):
    """
    Exception raised when the dataset is not part of the product.

    If you request a datasetID and a productID
    at the same time with the describe command,
    please verify that the dataset is part of the product.
    """

    def __init__(self, dataset_id: str, product_id: str):
        message = (
            f"{dataset_id} not part of {product_id} "
            f"Please check that the dataset is part of the product and "
            f"the input datasetID is correct."
        )
        super().__init__(message)


class ProductNotFound(Exception):
    """
    Exception raised when the product is not found in the catalogue.

    Possible reasons:

    - The productID is incorrect and not present in the catalogue.
    - The product has been retired.

    Please verify that the productID can be found in
    the result of the :func:`~copernicusmarine.describe` command.
    If yes, please contact the User Support, (widget chat on
    `Copernicus Marine website <https://help.marine.copernicus.eu/en/>`_).
    """

    def __init__(self, product_id: str):
        message = (
            f"{product_id} "
            f"Please check that the product exists and "
            f"the input productID is correct."
        )
        super().__init__(message)


REGEX_PATTERN_DATE_YYYYMM = r"[12]\d{3}(0[1-9]|1[0-2])"
PART_SEPARATOR = "--ext--"


def get_version_and_part_from_full_dataset_id(
    full_dataset_id: str,
) -> tuple[str, str, str]:
    if PART_SEPARATOR in full_dataset_id:
        name_with_maybe_version, part = full_dataset_id.split(PART_SEPARATOR)
    else:
        name_with_maybe_version = full_dataset_id
        part = PART_DEFAULT
    pattern = rf"^(.*?)(?:_({REGEX_PATTERN_DATE_YYYYMM}))?$"
    match = re.match(pattern, name_with_maybe_version)
    if match:
        dataset_name = match.group(1)
        version = match.group(2) or VERSION_DEFAULT
    else:
        raise Exception(f"Could not parse datasetID: {full_dataset_id}")
    return dataset_name, version, part



import fnmatch
import logging
import pathlib
import re
from dataclasses import dataclass, field
from datetime import datetime
from json import load
from typing import Any, Dict, List, Optional

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_FILE_FORMAT,
    DEFAULT_VERTICAL_AXIS,
    CoordinatesSelectionMethod,
    FileFormat,
    VerticalAxis,
)
from copernicusmarine.core_functions.utils import datetime_parser
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    TemporalParameters,
    XParameters,
    YParameters,
)

logger = logging.getLogger("copernicusmarine")


MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS: dict[str, str] = {
    "dataset_version": "force_dataset_version",
    "dataset_part": "force_dataset_part",
    "service": "force_service",
    "maximum_latitude": "maximum_y",
    "minimum_latitude": "minimum_y",
    "maximum_longitude": "maximum_x",
    "minimum_longitude": "minimum_x",
}


@dataclass
class SubsetRequest:
    dataset_id: str
    dataset_url: Optional[str] = None
    force_dataset_version: Optional[str] = None
    force_dataset_part: Optional[str] = None
    variables: Optional[List[str]] = None
    minimum_x: Optional[float] = None
    maximum_x: Optional[float] = None
    minimum_y: Optional[float] = None
    maximum_y: Optional[float] = None
    minimum_depth: Optional[float] = None
    maximum_depth: Optional[float] = None
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS
    start_datetime: Optional[datetime] = None
    end_datetime: Optional[datetime] = None
    platform_ids: Optional[List[str]] = None
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    )
    output_filename: Optional[str] = None
    file_format: FileFormat = DEFAULT_FILE_FORMAT
    force_service: Optional[str] = None
    output_directory: pathlib.Path = pathlib.Path(".")
    overwrite: bool = False
    skip_existing: bool = False
    netcdf_compression_level: int = 0
    netcdf3_compatible: bool = False
    dry_run: bool = False
    raise_if_updating: bool = False

    def update(self, new_dict: dict):
        """Method to update values in SubsetRequest object.
        Skips "None" values
        """
        for key, value in new_dict.items():
            if value is None or (
                isinstance(value, (list, tuple)) and len(value) < 1
            ):
                pass
            else:
                self.__dict__.update({key: value})

    def enforce_types(self):
        type_enforced_dict = {}
        for key, value in self.__dict__.items():
            if key in [
                "minimum_longitude",
                "maximum_longitude",
                "minimum_latitude",
                "maximum_latitude",
                "minimum_depth",
                "maximum_depth",
                "minimum_x",
                "maximum_x",
                "minimum_y",
                "maximum_y",
            ]:
                new_value = float(value) if value is not None else None
            elif key in [
                "start_datetime",
                "end_datetime",
            ]:
                new_value = datetime_parser(value) if value else None
            elif key in ["variables", "platform_ids"]:
                new_value = list(value) if value is not None else None
            elif key in ["output_directory"]:
                new_value = pathlib.Path(value) if value is not None else None
            else:
                new_value = str(value) if value else None
            type_enforced_dict[key] = new_value
        self.__dict__.update(type_enforced_dict)

    def from_file(self, filepath: pathlib.Path):
        json_file = open(filepath)
        json_content = load(json_file)

        json_with_deprecated_options_replace = {}

        for key, val in json_content.items():
            if key in DEPRECATED_OPTIONS:
                deprecated_option = DEPRECATED_OPTIONS[key]
                json_with_deprecated_options_replace[
                    deprecated_option.new_name
                ] = val
            elif key in MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS:
                new_key = MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS[key]
                json_with_deprecated_options_replace[new_key] = val
            else:
                json_with_deprecated_options_replace[key] = val

        self.__dict__.update(json_with_deprecated_options_replace)
        self.enforce_types()

    def get_temporal_parameters(
        self, axis_coordinate_id_mapping: dict[str, str]
    ) -> TemporalParameters:
        return TemporalParameters(
            start_datetime=self.start_datetime,
            end_datetime=self.end_datetime,
            coordinate_id=axis_coordinate_id_mapping.get("t", "time"),
        )

    def get_geographical_parameters(
        self,
        axis_coordinate_id_mapping: dict[str, str],
        is_original_grid: bool = False,
    ) -> GeographicalParameters:
        return GeographicalParameters(
            x_axis_parameters=XParameters(
                minimum_x=self.minimum_x,
                maximum_x=self.maximum_x,
                coordinate_id=axis_coordinate_id_mapping.get("x", "longitude"),
            ),
            y_axis_parameters=YParameters(
                minimum_y=self.minimum_y,
                maximum_y=self.maximum_y,
                coordinate_id=axis_coordinate_id_mapping.get("y", "latitude"),
            ),
            projection="originalGrid" if is_original_grid else "lonlat",
        )

    def get_depth_parameters(
        self, axis_coordinate_id_mapping: dict[str, str]
    ) -> DepthParameters:
        return DepthParameters(
            minimum_depth=self.minimum_depth,
            maximum_depth=self.maximum_depth,
            vertical_axis=self.vertical_axis,
            coordinate_id=axis_coordinate_id_mapping.get("z", "depth"),
        )


def convert_motu_api_request_to_structure(
    motu_api_request: str,
) -> SubsetRequest:
    prefix = "python -m motuclient "
    string = motu_api_request.replace(prefix, "").replace("'", "")
    arguments = [
        substr.strip() for substr in string.split("--")[1:]
    ]  # for subsubstr in substr.split(" ", maxsplit=1)]
    arg_value_tuples = [
        tuple(substr.split(" ", maxsplit=1)) for substr in arguments
    ]
    motu_api_request_dict: Dict[str, Any] = {}
    for arg, value in arg_value_tuples:
        if arg == "variable":
            # special case for variable, since it can have multiple values
            motu_api_request_dict.setdefault(arg, []).append(value)
        else:
            motu_api_request_dict[arg] = value
    subset_request = SubsetRequest(
        dataset_id="",
        output_directory=pathlib.Path("."),
        output_filename=None,
        force_service=None,
    )
    # TODO: I think i will need to change this to maximum_x and so
    conversion_dict = {
        "product-id": "dataset_id",
        "latitude-min": "minimum_latitude",
        "latitude-max": "maximum_latitude",
        "longitude-min": "minimum_longitude",
        "longitude-max": "maximum_longitude",
        "depth-min": "minimum_depth",
        "depth-max": "maximum_depth",
        "date-min": "start_datetime",
        "date-max": "end_datetime",
        "variable": "variables",
    }
    for key, value in motu_api_request_dict.items():
        if key in conversion_dict.keys():
            subset_request.__dict__.update({conversion_dict[key]: value})
    subset_request.enforce_types()
    return subset_request


@dataclass
class GetRequest:
    dataset_id: str
    dataset_url: Optional[str] = None
    force_dataset_version: Optional[str] = None
    force_dataset_part: Optional[str] = None
    no_directories: bool = False
    output_directory: str = "."
    overwrite: bool = False
    filter: Optional[str] = None
    regex: Optional[str] = None
    file_list: Optional[pathlib.Path] = None
    sync: bool = False
    sync_delete: bool = False
    index_parts: bool = False
    direct_download: Optional[list[str]] = None
    dry_run: bool = False
    skip_existing: bool = False

    def update(self, new_dict: dict):
        """Method to update values in GetRequest object.
        Skips "None" values
        """
        for key, value in new_dict.items():
            if value is None:
                pass
            else:
                self.__dict__.update({key: value})

    def enforce_types(self):
        type_enforced_dict = {}
        for key, value in self.__dict__.items():
            if key in [
                "no_directories",
                "sync",
                "sync_delete",
                "dry_run",
                "overwrite",
                "skip_existing",
                "index_parts",
            ]:
                new_value = bool(value) if value is not None else None
            else:
                new_value = str(value) if value else None
            type_enforced_dict[key] = new_value
        self.__dict__.update(type_enforced_dict)

    def from_file(self, filepath: pathlib.Path):
        json_file = load(open(filepath))
        json_with_mapped_options = {}
        for key, val in json_file.items():
            if key in MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS:
                new_key = MAPPING_REQUEST_FILES_AND_REQUEST_OPTIONS[key]
                json_with_mapped_options[new_key] = val
            else:
                json_with_mapped_options[key] = val
        self.__dict__.update(json_with_mapped_options)
        self.enforce_types()
        full_regex = self.regex
        if self.filter:
            filter_regex = filter_to_regex(self.filter)
            full_regex = overload_regex_with_additionnal_filter(
                filter_regex, full_regex
            )
        if self.file_list:
            file_list_regex = file_list_to_regex(self.file_list)
            full_regex = overload_regex_with_additionnal_filter(
                file_list_regex, full_regex
            )
        self.regex = full_regex


@dataclass
class LoadRequest:
    dataset_id: str
    dataset_url: Optional[str] = None
    force_dataset_version: Optional[str] = None
    force_dataset_part: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    variables: Optional[List[str]] = None
    platform_ids: Optional[List[str]] = None
    geographical_parameters: GeographicalParameters = field(
        default_factory=GeographicalParameters
    )
    temporal_parameters: TemporalParameters = field(
        default_factory=TemporalParameters
    )
    depth_parameters: DepthParameters = field(default_factory=DepthParameters)
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    )
    force_service: Optional[str] = None
    credentials_file: Optional[pathlib.Path] = None
    disable_progress_bar: bool = False

    def update_attributes(self, axis_coordinate_id_mapping: dict):
        self.geographical_parameters.x_axis_parameters.coordinate_id = (
            axis_coordinate_id_mapping.get("x", "")
        )
        self.geographical_parameters.y_axis_parameters.coordinate_id = (
            axis_coordinate_id_mapping.get("y", "")
        )
        self.temporal_parameters.coordinate_id = (
            axis_coordinate_id_mapping.get("t", "")
        )
        self.depth_parameters.coordinate_id = axis_coordinate_id_mapping.get(
            "z", ""
        )

    def to_subset_request(self) -> SubsetRequest:
        return SubsetRequest(
            dataset_id=self.dataset_id,
            dataset_url=self.dataset_url,
            force_dataset_version=self.force_dataset_version,
            force_dataset_part=self.force_dataset_part,
            variables=self.variables,
            platform_ids=self.platform_ids,
            minimum_x=self.geographical_parameters.x_axis_parameters.minimum_x,
            maximum_x=self.geographical_parameters.x_axis_parameters.maximum_x,
            minimum_y=self.geographical_parameters.y_axis_parameters.minimum_y,
            maximum_y=self.geographical_parameters.y_axis_parameters.maximum_y,
            minimum_depth=self.depth_parameters.minimum_depth,
            maximum_depth=self.depth_parameters.maximum_depth,
            vertical_axis=self.depth_parameters.vertical_axis,
            start_datetime=self.temporal_parameters.start_datetime,
            end_datetime=self.temporal_parameters.end_datetime,
            coordinates_selection_method=self.coordinates_selection_method,
            force_service=self.force_service,
        )


def filter_to_regex(filter: str) -> str:
    return fnmatch.translate(filter)


def file_list_to_regex(file_list_path: pathlib.Path) -> str:
    pattern = ""
    with open(file_list_path) as file_list:
        pattern = "|".join(map(re.escape, file_list.read().splitlines()))
    return pattern


def overload_regex_with_additionnal_filter(
    regex: str, filter: Optional[str]
) -> str:
    return "(" + regex + "|" + filter + ")" if filter else regex



from typing import Callable, Union

import pandas as pd
import xarray

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.credentials_utils import (
    get_and_check_username_password,
)
from copernicusmarine.core_functions.exceptions import (
    FormatNotSupported,
    ServiceNotSupported,
)
from copernicusmarine.core_functions.marine_datastore_config import (
    get_config_and_check_version_subset,
)
from copernicusmarine.core_functions.models import CommandType
from copernicusmarine.core_functions.request_structure import LoadRequest
from copernicusmarine.core_functions.services_utils import (
    RetrievalService,
    get_retrieval_service,
)
from copernicusmarine.download_functions.download_sparse import (
    read_dataframe_sparse,
)
from copernicusmarine.download_functions.download_zarr import (
    get_optimum_dask_chunking,
    rechunk,
)
from copernicusmarine.download_functions.subset_xarray import (
    check_dataset_subset_bounds,
    timestamp_or_datestring_to_datetime,
)


def load_data_object_from_load_request(
    load_request: LoadRequest,
    arco_series_load_function: Callable,
    chunks_factor_size_limit: int,
    command_type: CommandType,
) -> Union[xarray.Dataset, pd.DataFrame]:
    marine_datastore_config = get_config_and_check_version_subset(
        staging=False,
    )
    retrieval_service: RetrievalService = get_retrieval_service(
        dataset_id=load_request.dataset_id,
        force_dataset_version_label=load_request.force_dataset_version,
        force_dataset_part_label=load_request.force_dataset_part,
        force_service_name_or_short_name=load_request.force_service,
        command_type=command_type,
        dataset_subset=load_request.to_subset_request(),
        marine_datastore_config=marine_datastore_config,
    )
    username, password = get_and_check_username_password(
        load_request.username,
        load_request.password,
        load_request.credentials_file,
    )
    load_request.dataset_url = retrieval_service.uri
    check_dataset_subset_bounds(
        service=retrieval_service.service,
        part=retrieval_service.dataset_part,
        dataset_subset=load_request.to_subset_request(),
        coordinates_selection_method=load_request.coordinates_selection_method,
        axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
    )
    if (
        retrieval_service.service.service_format
        == CopernicusMarineServiceFormat.SQLITE
    ) and command_type == CommandType.OPEN_DATASET:
        raise FormatNotSupported(
            CopernicusMarineServiceFormat.SQLITE.value,
            command_type.value[0].value,
            CommandType.READ_DATAFRAME.value[0].value,
        )
    elif (
        retrieval_service.service.service_format
        == CopernicusMarineServiceFormat.SQLITE
    ) and command_type == CommandType.READ_DATAFRAME:
        return read_dataframe_sparse(
            username=username,
            subset_request=load_request.to_subset_request(),
            metadata_url=retrieval_service.metadata_url,
            service=retrieval_service.service,
            product_doi=retrieval_service.product_doi,
            disable_progress_bar=load_request.disable_progress_bar,
        )

    load_request.dataset_url = retrieval_service.uri
    check_dataset_subset_bounds(
        service=retrieval_service.service,
        part=retrieval_service.dataset_part,
        dataset_subset=load_request.to_subset_request(),
        coordinates_selection_method=load_request.coordinates_selection_method,
        axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
    )
    load_request.update_attributes(
        retrieval_service.axis_coordinate_id_mapping
    )
    if retrieval_service.service_name in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
    ]:
        if retrieval_service.dataset_valid_start_date:
            parsed_start_datetime = timestamp_or_datestring_to_datetime(
                retrieval_service.dataset_valid_start_date
            )
            if (
                not load_request.temporal_parameters.start_datetime
                or load_request.temporal_parameters.start_datetime
                < parsed_start_datetime
            ):
                load_request.temporal_parameters.start_datetime = (
                    parsed_start_datetime
                )
        if chunks_factor_size_limit > 0 and retrieval_service.dataset_chunking:
            optimum_dask_chunking = get_optimum_dask_chunking(
                service=retrieval_service.service,
                variables=load_request.variables,
                dataset_chunking=retrieval_service.dataset_chunking,
                chunk_size_limit=chunks_factor_size_limit,
                axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
            )
        else:
            optimum_dask_chunking = None
        dataset = arco_series_load_function(
            username=username,
            password=password,
            dataset_url=load_request.dataset_url,
            variables=load_request.variables,
            geographical_parameters=load_request.geographical_parameters,
            temporal_parameters=load_request.temporal_parameters,
            depth_parameters=load_request.depth_parameters,
            coordinates_selection_method=load_request.coordinates_selection_method,
            opening_dask_chunks=optimum_dask_chunking,
        )
        if isinstance(dataset, xarray.Dataset):
            dataset = rechunk(
                dataset, optimum_dask_chunking, chunks_factor_size_limit
            )
    else:
        raise ServiceNotSupported(retrieval_service.service_name)
    return dataset



import logging
from typing import Optional

from copernicusmarine.catalogue_parser.catalogue_parser import (
    filter_catalogue_with_strings,
    merge_catalogues,
    parse_catalogue,
)
from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineCatalogue,
    DatasetNotFound,
    ProductNotFound,
)
from copernicusmarine.core_functions.marine_datastore_config import (
    get_config_and_check_version_describe,
)

logger = logging.getLogger("copernicusmarine")


def describe_function(
    show_all_versions: bool,
    contains: list[str],
    force_product_id: Optional[str],
    force_dataset_id: Optional[str],
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    staging: bool,
) -> CopernicusMarineCatalogue:
    marine_datasetore_config = get_config_and_check_version_describe(staging)
    if staging:
        logger.warning(
            "Detecting staging flag for describe command. "
            "Data will come from the staging environment."
        )
    catalogues: list[CopernicusMarineCatalogue] = []
    for catalogue_config in marine_datasetore_config.catalogues:
        try:
            catalogue: CopernicusMarineCatalogue = parse_catalogue(
                force_product_id=force_product_id,
                force_dataset_id=force_dataset_id,
                max_concurrent_requests=max_concurrent_requests,
                disable_progress_bar=disable_progress_bar,
                catalogue_config=catalogue_config,
            )
            catalogues.append(catalogue)
        except DatasetNotFound:
            logger.debug(
                f"Dataset not found in catalogue {catalogue_config.root_metadata_url}"
            )

    # Merge all catalogues
    base_catalogue = merge_catalogues(catalogues)
    if (force_dataset_id or force_product_id) and not base_catalogue.products:
        if force_product_id:
            raise ProductNotFound(
                product_id=force_product_id,
            )
        if force_dataset_id:
            raise DatasetNotFound(
                dataset_id=force_dataset_id,
            )
    if not show_all_versions:
        base_catalogue.filter_only_official_versions_and_parts()

    response_catalogue = (
        filter_catalogue_with_strings(base_catalogue, set(contains))
        if contains
        else base_catalogue
    )
    return response_catalogue



import logging
import ssl
from typing import Any, Literal, Optional

import boto3
import botocore
import botocore.config
import certifi
import requests
import requests.auth
from requests.adapters import HTTPAdapter, Retry

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_DISABLE_SSL_CONTEXT,
    COPERNICUSMARINE_HTTPS_RETRIES,
    COPERNICUSMARINE_HTTPS_TIMEOUT,
    COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH,
    COPERNICUSMARINE_TRUST_ENV,
    PROXY_HTTP,
    PROXY_HTTPS,
)
from copernicusmarine.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
    create_custom_query_function,
)

logger = logging.getLogger("copernicusmarine")

TRUST_ENV = COPERNICUSMARINE_TRUST_ENV == "True"
PROXIES = {}
if PROXY_HTTP:
    PROXIES["http"] = PROXY_HTTP
if PROXY_HTTPS:
    PROXIES["https"] = PROXY_HTTPS
try:
    HTTPS_TIMEOUT = float(COPERNICUSMARINE_HTTPS_TIMEOUT)
except ValueError:
    HTTPS_TIMEOUT = 60
try:
    HTTPS_RETRIES = int(COPERNICUSMARINE_HTTPS_RETRIES)
except ValueError:
    HTTPS_RETRIES = 5


def get_ssl_context() -> Optional[ssl.SSLContext]:
    if COPERNICUSMARINE_DISABLE_SSL_CONTEXT == "True":
        return None
    if COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH:
        return ssl.create_default_context(
            capath=COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH
        )
    return ssl.create_default_context(cafile=certifi.where())


def get_configured_boto3_session(
    endpoint_url: str,
    operation_type: list[Literal["ListObjectsV2", "HeadObject", "GetObject"]],
    username: Optional[str] = None,
    return_ressources: bool = False,
) -> tuple[Any, Any]:
    config_boto3 = botocore.config.Config(
        # s3={"addressing_style": "virtual"},
        signature_version=botocore.UNSIGNED,
        retries={"max_attempts": 10, "mode": "standard"},
    )
    s3_session = boto3.Session()
    s3_client = s3_session.client(
        "s3",
        config=config_boto3,
        endpoint_url=endpoint_url,
    )
    for operation in operation_type:
        # Register the botocore event handler for adding custom query params
        # to S3 HEAD and GET requests
        s3_client.meta.events.register(
            f"before-call.s3.{operation}",
            create_custom_query_function(username),
        )
    if not return_ressources:
        return s3_client, None
    s3_resource = boto3.resource(
        "s3",
        config=config_boto3,
        endpoint_url=endpoint_url,
    )
    return s3_client, s3_resource


# TODO: add tests
# example: with https://httpbin.org/delay/10 or
# https://medium.com/@mpuig/testing-robust-requests-with-python-a06537d97771
class ConfiguredRequestsSession(requests.Session):
    def __init__(
        self,
        *args,
        timeout: float = HTTPS_TIMEOUT,
        retries: int = HTTPS_RETRIES,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.trust_env = TRUST_ENV
        if COPERNICUSMARINE_DISABLE_SSL_CONTEXT == "True":
            self.verify = False
        else:
            self.verify = (
                COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH or certifi.where()
            )
        self.proxies = PROXIES
        if retries:
            self.mount(
                "https://",
                HTTPAdapter(
                    max_retries=Retry(
                        total=retries,
                        backoff_factor=1,
                        status_forcelist=[408, 429, 500, 502, 503, 504],
                    )
                ),
            )
        self.timeout = timeout

    def request(self, *args, **kwargs):
        kwargs.setdefault("timeout", self.timeout)
        return super().request(*args, **kwargs)


def get_configured_requests_session(
    timeout: float = HTTPS_TIMEOUT, retries: int = HTTPS_RETRIES
) -> requests.Session:
    return ConfiguredRequestsSession(timeout=timeout, retries=retries)


# from https://stackoverflow.com/questions/29931671/making-an-api-call-in-python-with-an-api-that-requires-a-bearer-token # noqa
class BearerAuth(requests.auth.AuthBase):
    """
    Allow to pass the bearer as "auth" argument to the requests.get method and not as a header.

    Hence the headers are not overwritten by the netrc file as stated here:
    https://requests.readthedocs.io/en/latest/user/authentication/#netrc-authentication
    """  # noqa

    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r


class JsonParserConnection:
    def __init__(
        self, timeout: float = HTTPS_TIMEOUT, retries: int = HTTPS_RETRIES
    ) -> None:
        self.session = get_configured_requests_session(timeout, retries)

    def get_json_file(self, url: str) -> dict[str, Any]:
        logger.debug(f"Fetching json file at this url: {url}")
        with self.session.get(
            url,
            params=construct_query_params_for_marine_data_store_monitoring(),
            proxies=self.session.proxies,
        ) as response:
            response.raise_for_status()
            return response.json()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.session.close()



from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from copernicusmarine.core_functions.models import (
    DEFAULT_GEOSPATIAL_PROJECTION,
    DEFAULT_VERTICAL_AXIS,
    GeoSpatialProjection,
    VerticalAxis,
)


@dataclass
class YParameters:
    minimum_y: Optional[float] = None
    maximum_y: Optional[float] = None
    coordinate_id: str = "latitude"


@dataclass
class XParameters:
    minimum_x: Optional[float] = None
    maximum_x: Optional[float] = None
    coordinate_id: str = "longitude"


@dataclass
class GeographicalParameters:
    y_axis_parameters: YParameters = field(default_factory=YParameters)
    x_axis_parameters: XParameters = field(default_factory=XParameters)
    projection: GeoSpatialProjection = DEFAULT_GEOSPATIAL_PROJECTION


@dataclass
class TemporalParameters:
    start_datetime: Optional[datetime] = None
    end_datetime: Optional[datetime] = None
    coordinate_id: str = "time"


@dataclass
class DepthParameters:
    minimum_depth: Optional[float] = None
    maximum_depth: Optional[float] = None
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS
    coordinate_id: str = "depth"
